# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for reversible layers."""

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from trax import fastmath
import trax.layers as tl


BACKENDS = [fastmath.Backend.JAX]


class ReversibleLayerTest(parameterized.TestCase):

  @parameterized.named_parameters([('_' + b.value, b) for b in BACKENDS])
  def test_reversible_swap(self, backend):
    with fastmath.use_backend(backend):
      layer = tl.ReversibleSwap()
      xs = [np.array([1, 2]), np.array([10, 20])]
      ys = layer(xs)
      self.assertEqual(tl.to_list(ys), [[10, 20], [1, 2]])


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for RNNs."""

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from trax import fastmath
from trax import shapes
from trax.models import rnn

BACKENDS = [fastmath.Backend.JAX]


@parameterized.named_parameters(
    ('_' + b.value, b) for b in BACKENDS)
class RNNTest(parameterized.TestCase):

  def test_rnnlm_forward_shape(self, backend):
    with fastmath.use_backend(backend):
      model = rnn.RNNLM(vocab_size=20, d_model=16)
      x = np.ones((3, 28)).astype(np.int32)
      _, _ = model.init(shapes.signature(x))
      y = model(x)
      self.assertEqual(y.shape, (3, 28, 20))

  def test_grulm_forward_shape(self, backend):
    with fastmath.use_backend(backend):
      model = rnn.GRULM(vocab_size=20, d_model=16)
      x = np.ones((3, 28)).astype(np.int32)
      _, _ = model.init(shapes.signature(x))
      y = model(x)
      self.assertEqual(y.shape, (3, 28, 20))

  def test_lstmseq2seqattn_forward_shape(self, backend):
    with fastmath.use_backend(backend):
      model = rnn.LSTMSeq2SeqAttn(
          input_vocab_size=20, target_vocab_size=20, d_model=16)
      x = np.ones((3, 28)).astype(np.int32)
      _, _ = model.init([shapes.signature(x), shapes.signature(x)])
      ys = model([x, x])
      self.assertEqual([y.shape for y in ys], [(3, 28, 20), (3, 28)])


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for RL training."""

import os
from absl.testing import absltest
import gym
import numpy as np
from trax import test_utils
from trax.rl import task as rl_task


class DummyEnv:
  """Dummy Env class for testing."""

  observation_space = gym.spaces.Box(-2, 2, shape=(2,))
  action_space = gym.spaces.Discrete(2)

  def reset(self):
    self._step = 0
    return np.ones((2,))

  def step(self, action):
    del action
    info = {
        'control_mask': self._step % 2 == 0,
        'discount_mask': self._step % 3 == 0,
    }
    self._step += 1
    return np.ones((2,)), 0.0, False, info


class TaskTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    test_utils.ensure_flag('test_tmpdir')

  def _extend(
      self, trajectory, action=0, dist_inputs=0, reward=0, done=False,
      new_observation=0,
  ):
    trajectory.extend(
        action=action, dist_inputs=dist_inputs, reward=reward, done=done,
        new_observation=new_observation,
    )

  def test_trajectory_len(self):
    """Test that trajectory length is equal to the number of observations."""
    tr = rl_task.Trajectory(observation=0)
    for _ in range(5):
      self._extend(tr)
    self.assertLen(tr, 6)

  def test_empty_trajectory_last_observation(self):
    """Test that last_observation is the one passed in __init__."""
    tr = rl_task.Trajectory(observation=123)
    self.assertEqual(tr.last_observation, 123)

  def test_nonempty_trajectory_last_observation(self):
    """Test that last_observation is the one passed in the last extend()."""
    tr = rl_task.Trajectory(observation=123)
    for _ in range(5):
      self._extend(tr)
    self._extend(tr, new_observation=321)
    self.assertEqual(tr.last_observation, 321)

  def test_trajectory_done_get_and_set(self):
    """Test that we can get and set the `done` flag of a trajectory."""
    tr = rl_task.Trajectory(observation=123)
    self._extend(tr)
    self.assertFalse(tr.done)
    tr.done = True
    self.assertTrue(tr.done)

  def test_trajectory_suffix_len(self):
    """Test that a trajectory suffix has the correct length."""
    tr = rl_task.Trajectory(observation=0)
    for _ in range(5):
      self._extend(tr)
    tr_suffix = tr.suffix(length=3)
    self.assertLen(tr_suffix, 3)

  def test_trajectory_suffix_observations(self):
    """Test that a trajectory suffix has the correct observations."""
    tr = rl_task.Trajectory(observation=0)
    for obs in range(1, 6):
      self._extend(tr, new_observation=obs)
    tr_suffix = tr.suffix(length=4)
    self.assertEqual([ts.observation for ts in tr_suffix.timesteps], [2, 3, 4])
    self.assertEqual(tr_suffix.last_observation, 5)

  def test_trajectory_to_np_shape(self):
    """Test that the shape of a to_np result matches the trajectory length."""
    tr = rl_task.Trajectory(observation=np.zeros((2, 3)))
    for _ in range(5):
      self._extend(tr, new_observation=np.zeros((2, 3)))
    tr_np = tr.to_np()
    self.assertEqual(tr_np.observation.shape, (len(tr), 2, 3))
    self.assertEqual(tr_np.action.shape, (len(tr),))

  def test_trajectory_to_np_shape_after_extend(self):
    """Test that the shape of a to_np result grows after calling extend()."""
    tr = rl_task.Trajectory(observation=0)
    for _ in range(5):
      self._extend(tr)
    len_before = tr.to_np().observation.shape[0]
    self._extend(tr)
    len_after = tr.to_np().observation.shape[0]
    self.assertEqual(len_after, len_before + 1)

  def test_trajectory_to_np_observations(self):
    """Test that to_np returns correct observations."""
    tr = rl_task.Trajectory(observation=0)
    for obs in range(1, 3):
      self._extend(tr, new_observation=obs)
    tr_np = tr.to_np()
    np.testing.assert_array_equal(tr_np.observation, [0, 1, 2])

  def test_trajectory_to_np_adds_margin(self):
    """Test that to_np adds a specified margin."""
    tr = rl_task.Trajectory(observation=2)
    for _ in range(2):
      self._extend(tr, new_observation=2)
    tr_np = tr.to_np(margin=2)
    np.testing.assert_array_equal(tr_np.observation, [2, 2, 2, 0])
    np.testing.assert_array_equal(tr_np.mask, [1, 1, 0, 0])

  def test_trajectory_to_np_without_margin_cuts_last_observation(self):
    """Test that to_np with margin=0 cuts the last observation."""
    tr = rl_task.Trajectory(observation=0)
    for obs in range(1, 4):
      self._extend(tr, new_observation=obs)
    tr_np = tr.to_np(margin=0)
    np.testing.assert_array_equal(tr_np.observation, [0, 1, 2])

  def test_task_random_initial_trajectories_and_max_steps(self):
    """Test generating initial random trajectories, stop at max steps."""
    task = rl_task.RLTask(DummyEnv(), initial_trajectories=1, max_steps=9)
    stream = task.trajectory_slice_stream(max_slice_length=1)
    next_slice = next(stream)
    self.assertEqual(next_slice.observation.shape, (1, 2))

  def test_time_limit_terminates_epsiodes(self):
    """Test that episodes are terminated upon reaching `time_limit` steps."""
    task = rl_task.RLTask(
        DummyEnv(), initial_trajectories=3, max_steps=10, time_limit=10
    )
    trajectories = task.trajectories[0]  # Get trajectories from epoch 0.
    self.assertLen(trajectories, 3)
    for trajectory in trajectories:
      self.assertTrue(trajectory.done)
      # max_steps + 1 (the initial observation doesn't count).
      self.assertLen(trajectory, 11)

  def test_max_steps_doesnt_terminate_epsiodes(self):
    """Test that episodes are not terminated upon reaching `max_steps` steps."""
    task = rl_task.RLTask(
        DummyEnv(), initial_trajectories=2, max_steps=5, time_limit=10
    )
    trajectories = task.trajectories[0]  # Get trajectories from epoch 0.
    self.assertLen(trajectories, 2)
    # The trajectory should be cut in half. The first half should not be "done".
    self.assertFalse(trajectories[0].done)
    self.assertLen(trajectories[0], 6)  # max_steps + 1
    # The second half should be "done".
    self.assertTrue(trajectories[1].done)
    self.assertLen(trajectories[1], 6)  # max_steps + 1

  def test_collects_specified_number_of_interactions(self):
    """Test that the specified number of interactions are collected."""
    task = rl_task.RLTask(
        DummyEnv(), initial_trajectories=0, max_steps=3, time_limit=20
    )
    task.collect_trajectories(policy=(lambda _: (0, 0)), n_interactions=10)
    trajectories = task.trajectories[1]  # Get trajectories from epoch 1.
    n_interactions = 0
    for trajectory in trajectories:
      n_interactions += len(trajectory) - 1
    self.assertEqual(n_interactions, 10)

  def test_collects_specified_number_of_trajectories(self):
    """Test that the specified number of interactions are collected."""
    task = rl_task.RLTask(
        DummyEnv(), initial_trajectories=0, max_steps=3, time_limit=20
    )
    task.collect_trajectories(policy=(lambda _: (0, 0)), n_trajectories=3)
    trajectories = task.trajectories[1]  # Get trajectories from epoch 1.
    self.assertLen(trajectories, 3)

  def test_task_save_init(self):
    """Test saving and re-initialization."""
    task1 = rl_task.RLTask(DummyEnv(), initial_trajectories=13,
                           max_steps=9, gamma=0.9)
    self.assertLen(task1.trajectories[0], 13)
    self.assertEqual(task1.max_steps, 9)
    self.assertEqual(task1.gamma, 0.9)
    temp_file = os.path.join(self.create_tempdir().full_path, 'task.pkl')
    task1.save_to_file(temp_file)
    task2 = rl_task.RLTask(DummyEnv(), initial_trajectories=3,
                           max_steps=19, gamma=1.0)
    self.assertLen(task2.trajectories[0], 3)
    self.assertEqual(task2.max_steps, 19)
    self.assertEqual(task2.gamma, 1.0)
    task2.init_from_file(temp_file)
    self.assertLen(task2.trajectories[0], 13)
    self.assertEqual(task2.max_steps, 9)
    self.assertEqual(task2.gamma, 0.9)

  def test_task_epochs_index_minusone(self):
    """Test that the epoch index -1 means last epoch and updates to it."""
    obs = np.zeros((2,))
    tr1 = rl_task.Trajectory(obs)
    self._extend(tr1, new_observation=obs, done=True)
    task = rl_task.RLTask(DummyEnv(), initial_trajectories=[tr1], max_steps=9)
    stream = task.trajectory_slice_stream(epochs=[-1], max_slice_length=1)
    next_slice = next(stream)
    np.testing.assert_equal(next_slice.observation, np.zeros((1, 2)))
    task.collect_trajectories(policy=(lambda _: (0, 0)), n_trajectories=1)
    next_slice = next(stream)
    np.testing.assert_equal(next_slice.observation, np.ones((1, 2)))

  def test_trajectory_slice_stream_shape(self):
    """Test the shape yielded by trajectory stream."""
    obs = np.zeros((12, 13))
    tr1 = rl_task.Trajectory(obs)
    self._extend(tr1, new_observation=obs, done=True)
    task = rl_task.RLTask(DummyEnv(), initial_trajectories=[tr1], max_steps=9)
    stream = task.trajectory_slice_stream(max_slice_length=1)
    next_slice = next(stream)
    self.assertEqual(next_slice.observation.shape, (1, 12, 13))

  def test_trajectory_slice_stream_long_slice(self):
    """Test trajectory stream with slices of longer length."""
    obs = np.zeros((12, 13))
    tr1 = rl_task.Trajectory(obs)
    self._extend(tr1, new_observation=obs)
    self._extend(tr1, new_observation=obs, done=True)
    task = rl_task.RLTask(DummyEnv(), initial_trajectories=[tr1], max_steps=9)
    stream = task.trajectory_slice_stream(max_slice_length=2)
    next_slice = next(stream)
    self.assertEqual(next_slice.observation.shape, (2, 12, 13))

  def test_trajectory_slice_stream_sampling_uniform(self):
    """Test if the trajectory stream samples uniformly."""
    # Long trajectory of 0s.
    tr1 = rl_task.Trajectory(0)
    for _ in range(100):
      self._extend(tr1)
    self._extend(tr1, new_observation=200, done=True)
    # Short trajectory of 101.
    tr2 = rl_task.Trajectory(101)
    self._extend(tr2, new_observation=200, done=True)
    task = rl_task.RLTask(
        DummyEnv(), initial_trajectories=[tr1, tr2], max_steps=9)

    # Stream of both. Check that we're sampling by slice, not by trajectory.
    stream = task.trajectory_slice_stream(max_slice_length=1)
    slices = []
    for _ in range(10):
      next_slice = next(stream)
      assert next_slice.observation.shape[0] == 1
      slices.append(next_slice.observation[-1])
    mean_obs = sum(slices) / float(len(slices))
    # Average should be around 1 sampling from 0x100, 101 uniformly.
    self.assertLess(mean_obs, 31)  # Sampling 101 even 3 times is unlikely.
    self.assertLen(slices, 10)

  def test_trajectory_slice_stream_sampling_by_trajectory(self):
    """Test if the trajectory stream samples by trajectory."""
    # Long trajectory of 0s.
    tr1 = rl_task.Trajectory(0)
    for _ in range(100):
      self._extend(tr1)
    self._extend(tr1, new_observation=200, done=True)
    # Short trajectory of 101.
    tr2 = rl_task.Trajectory(101)
    self._extend(tr2, new_observation=200, done=True)
    task = rl_task.RLTask(
        DummyEnv(), initial_trajectories=[tr1, tr2], max_steps=9)

    # Stream of both. Check that we're sampling by trajectory.
    stream = task.trajectory_slice_stream(
        max_slice_length=1, sample_trajectories_uniformly=True)
    slices = []
    for _ in range(10):
      next_slice = next(stream)
      assert next_slice.observation.shape[0] == 1
      slices.append(next_slice.observation[-1])
    mean_obs = sum(slices) / float(len(slices))
    # Average should be around 50, sampling from {0, 101} uniformly.
    # Sampling 101 < 2 times has low probability (but it possible, flaky test).
    self.assertGreater(mean_obs, 20)
    self.assertLen(slices, 10)

  def test_trajectory_slice_stream_margin(self):
    """Test trajectory stream with an added margin."""
    tr1 = rl_task.Trajectory(0)
    self._extend(tr1, new_observation=1)
    self._extend(tr1, new_observation=1)
    self._extend(
        tr1, new_observation=1, action=1, dist_inputs=2, reward=3, done=True
    )
    task = rl_task.RLTask(DummyEnv(), initial_trajectories=[tr1], max_steps=9)

    # Stream of slices without the final state.
    stream1 = task.trajectory_slice_stream(max_slice_length=4, margin=3)
    got_done = False
    for _ in range(20):
      next_slice = next(stream1)
      self.assertEqual(next_slice.observation.shape, (4,))
      if next_slice.done[0]:
        # In the slice, first we have the last timestep in the actual
        # trajectory, so observation = 1.
        # Then comes the first timestep in the margin, which has the final
        # observation from the trajectory: observation = 1.
        # The remaining timesteps have 0 observations.
        np.testing.assert_array_equal(next_slice.observation, [1, 1, 0, 0])
        # In the margin, done = True and mask = 0.
        for i in range(1, next_slice.observation.shape[0]):
          self.assertTrue(next_slice.done[i])
          self.assertFalse(next_slice.mask[i])
        got_done = True
    # Assert that we got a done somewhere, otherwise the test is not triggered.
    # Not getting done has low probability (1/2^20) but is possible, flaky test.
    self.assertTrue(got_done)

  def test_trajectory_batch_stream_propagates_env_info(self):
    task = rl_task.RLTask(DummyEnv(), initial_trajectories=1, max_steps=4)
    stream = task.trajectory_batch_stream(batch_size=1, max_slice_length=4)
    tr_slice = next(stream)
    # control_mask = step % 2 == 0, discount_mask = step % 3 == 0.
    np.testing.assert_array_equal(
        tr_slice.env_info.control_mask, [[1, 0, 1, 0]]
    )
    np.testing.assert_array_equal(
        tr_slice.env_info.discount_mask, [[1, 0, 0, 1]]
    )

  def test_trajectory_batch_stream_shape(self):
    task = rl_task.RLTask(DummyEnv(), initial_trajectories=1, max_steps=10)
    batch_stream = task.trajectory_batch_stream(
        batch_size=3, min_slice_length=4, max_slice_length=4
    )
    batch = next(batch_stream)
    self.assertEqual(batch.observation.shape, (3, 4, 2))


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions and classes for obtaining and preprocesing data.

The ``trax.data`` module presents a flattened (no subpackages) public API.
(Many of the functions and class initilizers in the API are also accessible for
gin configuration.) To use as a client, import ``trax.data`` and access
functions using ``data.foo`` qualified names; for example::

   from trax import data
   ...
   training_inputs = data.Serial(
     ...
     data.Tokenize(),
     data.Shuffle(),
     ...
  )

"""

from trax.data.debug_data_pipeline import debug_pipeline

from trax.data.inputs import add_loss_weights
from trax.data.inputs import addition_inputs
from trax.data.inputs import AddLossWeights
from trax.data.inputs import AppendValue
from trax.data.inputs import batch
from trax.data.inputs import Batch
from trax.data.inputs import bucket_by_length
from trax.data.inputs import BucketByLength
from trax.data.inputs import CastTo
from trax.data.inputs import ConcatenateToLMInput
from trax.data.inputs import consume_noise_mask
from trax.data.inputs import CountAndSkip
from trax.data.inputs import Dup
from trax.data.inputs import FilterByLength
from trax.data.inputs import FilterEmptyExamples
from trax.data.inputs import generate_random_noise_mask
from trax.data.inputs import generate_sequential_chunks
from trax.data.inputs import Log
from trax.data.inputs import MLM
from trax.data.inputs import PadToLength
from trax.data.inputs import Parallel
from trax.data.inputs import Prefetch
from trax.data.inputs import PrefixLM
from trax.data.inputs import random_spans_noise_mask
from trax.data.inputs import sequence_copy_inputs
from trax.data.inputs import Serial
from trax.data.inputs import shuffle
from trax.data.inputs import Shuffle
from trax.data.inputs import simple_sequence_copy_inputs
from trax.data.inputs import sine_inputs
from trax.data.inputs import TruncateToLength
from trax.data.inputs import UnBatch
from trax.data.inputs import UniformlySeek

from trax.data.tf_inputs import add_eos_to_output_features
from trax.data.tf_inputs import BertGlueEvalStream
from trax.data.tf_inputs import BertGlueTrainStream
from trax.data.tf_inputs import BertNextSentencePredictionInputs
from trax.data.tf_inputs import cifar10_augmentation_flatten_preprocess
from trax.data.tf_inputs import cifar10_augmentation_preprocess
from trax.data.tf_inputs import ConvertToUnicode
from trax.data.tf_inputs import CorpusToRandomChunks
from trax.data.tf_inputs import CreateAnnotatedDropInputs
from trax.data.tf_inputs import CreateAquaInputs
from trax.data.tf_inputs import CreateBertInputs
from trax.data.tf_inputs import CreateDropInputs
from trax.data.tf_inputs import CreateMathQAInputs
from trax.data.tf_inputs import data_streams
from trax.data.tf_inputs import detokenize
from trax.data.tf_inputs import downsampled_imagenet_flatten_bare_preprocess
from trax.data.tf_inputs import filter_dataset_on_len
from trax.data.tf_inputs import lm1b_preprocess
from trax.data.tf_inputs import mask_random_tokens
from trax.data.tf_inputs import sentencepiece_tokenize
from trax.data.tf_inputs import SentencePieceTokenize
from trax.data.tf_inputs import squeeze_targets_preprocess
from trax.data.tf_inputs import T5GlueEvalStream
from trax.data.tf_inputs import T5GlueEvalStreamsParallel
from trax.data.tf_inputs import T5GlueEvalTasks
from trax.data.tf_inputs import T5GlueTrainStream
from trax.data.tf_inputs import T5GlueTrainStreamsParallel
from trax.data.tf_inputs import TFDS
from trax.data.tf_inputs import tokenize
from trax.data.tf_inputs import Tokenize
from trax.data.tf_inputs import truncate_dataset_on_len
from trax.data.tf_inputs import vocab_size
from trax.data.tf_inputs import wmt_concat_preprocess
from trax.data.tf_inputs import wmt_preprocess




# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Optimizers for use with Trax layers."""

import gin

from trax.optimizers import adafactor
from trax.optimizers import adam
from trax.optimizers import base
from trax.optimizers import momentum
from trax.optimizers import rms_prop
from trax.optimizers import sm3
from trax.optimizers import trainer
from trax.optimizers.trainer import ReversibleSerialTrainer
from trax.optimizers.trainer import Trainer


def opt_configure(*args, **kwargs):
  kwargs['module'] = 'trax.optimizers'
  return gin.external_configurable(*args, **kwargs)

# Optimizers (using upper-case names).
# pylint: disable=invalid-name
SGD = opt_configure(base.SGD)
Momentum = opt_configure(momentum.Momentum)
RMSProp = opt_configure(rms_prop.RMSProp)
Adam = opt_configure(adam.Adam)
Adafactor = opt_configure(adafactor.Adafactor)
SM3 = opt_configure(sm3.SM3)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Trax Transpose Convolution Layers."""

import functools
import itertools
import operator

from jax import lax
from trax import fastmath
from trax.fastmath import numpy as jnp
from trax.layers import base
from trax.layers import initializers as init


class ConvTranspose(base.Layer):
  """Layer constructor function for a general Transpose Convolutional Layer."""

  def __init__(self,
               filters,
               kernel_size,
               strides=None,
               padding='VALID',
               rhs_dilation=None,
               dimension_numbers=('NHWC', 'HWIO', 'NHWC'),
               kernel_initialzer=None,
               bias_initializer=init.RandomNormalInitializer(1e-6)):
    super(ConvTranspose, self).__init__()
    self._filters = filters
    self._kernel_size = kernel_size
    self._padding = padding
    self._rhs_dilation = rhs_dilation
    self._dimension_numbers = dimension_numbers
    self._lhs_spec, self._rhs_spec, self._out_spec = dimension_numbers
    self._one = (1,) * len(kernel_size)
    self._strides = strides or self._one
    self._bias_initializer = bias_initializer
    rhs_spec = self._rhs_spec
    self._kernel_initializer = kernel_initialzer
    if kernel_initialzer is None:
      self._kernel_initializer = init.GlorotNormalInitializer(
          rhs_spec.index('O'), rhs_spec.index('I'))

  def _check_nhwc(self):
    msg = 'Deconvolutions on more than 4 dimensions only supported in NHWC.'
    assert self._lhs_spec == self._out_spec == 'NHWC', msg

  def forward(self, x):
    w, b = self.weights
    x_shape = list(x.shape)
    if len(x_shape) > 4:
      self._check_nhwc()
      new_batch_dim = functools.reduce(operator.mul, x.shape[:-3])
      x = jnp.reshape(x, [new_batch_dim] + list(x.shape[-3:]))
    res = lax.conv_transpose(x, w, self._strides, self._padding,
                             self._rhs_dilation, self._dimension_numbers) + b
    if len(x_shape) > 4:
      res = jnp.reshape(res, x_shape[:-3] + list(res.shape[-3:]))
    return res

  def _kernel_shape(self, input_shape):
    """Helper to calculate the kernel shape."""
    kernel_size_iter = iter(self._kernel_size)
    return [
        self._filters if c == 'O' else input_shape[self._lhs_spec.index('C')]
        if c == 'I' else next(kernel_size_iter) for c in self._rhs_spec
    ]

  def init_weights_and_state(self, input_signature):
    input_shape = input_signature.shape
    if len(input_shape) > 4:
      self._check_nhwc()
      new_batch_dim = functools.reduce(operator.mul, input_shape[:-3])
      input_shape = [new_batch_dim] + list(input_shape[-3:])
    kernel_shape = self._kernel_shape(input_shape)
    bias_shape = [self._filters if c == 'C' else 1 for c in self._out_spec]
    bias_shape = tuple(itertools.dropwhile(lambda x: x == 1, bias_shape))
    rng1, rng2 = fastmath.random.split(self.rng, 2)
    w = self._kernel_initializer(kernel_shape, rng1)
    b = self._bias_initializer(bias_shape, rng2)
    self.weights = (w, b)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Utilities for serializing trajectories into discrete sequences."""

import functools

import gym
import numpy as np

from trax import layers as tl
from trax.fastmath import numpy as jnp
from trax.rl import space_serializer


# pylint: disable=invalid-name
# TODO(pkozakowski): Move the layers to trax.layers and remove this module.
def Serialize(serializer):
  """Layer that serializes a given array."""
  def serialize(x):
    (batch_size, length) = x.shape[:2]
    shape_suffix = x.shape[2:]
    x = jnp.reshape(x, (batch_size * length,) + shape_suffix)
    x = serializer.serialize(x)
    return jnp.reshape(x, (batch_size, -1, serializer.representation_length,))
  return tl.Fn('Serialize', serialize)


def Interleave():
  """Layer that interleaves and flattens two serialized sequences.

  The first sequence can be longer by 1 than the second one. This is so we can
  interleave sequences of observations and actions, when there's 1 extra
  observation at the end.

  For serialized sequences [[x_1_1, ..., x_1_R1], ..., [x_L1_1, ..., x_L1_R1]]
  and [[y_1_1, ..., y_1_R2], ..., [y_L2_1, ..., y_L2_R2]], where L1 = L2 + 1,
  the result is [x_1_1, ..., x_1_R1, y_1_1, ..., y_1_R2, ..., x_L2_1, ...,
  x_L2_R1, y_L2_1, ..., y_L2_R2, x_L1_1, ..., x_L1_R1] (batch dimension omitted
  for clarity).

  The layer inputs are a sequence pair of shapes (B, L1, R1) and (B, L2, R2),
  where B is batch size, L* is the length of the sequence and R* is the
  representation length of each element in the sequence.

  Returns:
    Layer that interleaves sequence of shape (B, L1 * R1 + L2 * R2).
  """
  def interleave(x, y):
    (batch_size, _, _) = x.shape
    (_, length, _) = y.shape
    assert x.shape[1] in (length, length + 1)

    reprs = jnp.concatenate((x[:, :length], y), axis=2)
    reprs = jnp.reshape(reprs, (batch_size, -1))
    remainder = jnp.reshape(x[:, length:], (batch_size, -1))
    return jnp.concatenate((reprs, remainder), axis=1)
  return tl.Fn('Interleave', interleave)


def Deinterleave(x_size, y_size):
  """Layer that does the inverse of Interleave."""
  def deinterleave(inputs):
    reprs = inputs
    (batch_size, length) = reprs.shape[:2]
    shape_suffix = reprs.shape[2:]
    remainder_length = length % (x_size + y_size)
    if remainder_length > 0:
      remainder = reprs[:, None, -remainder_length:]
      reprs = reprs[:, :-remainder_length]
    reprs = jnp.reshape(reprs, (batch_size, -1, x_size + y_size) + shape_suffix)
    x_reprs = reprs[:, :, :x_size]
    y_reprs = reprs[:, :, x_size:]
    if remainder_length > 0:
      x_reprs = jnp.concatenate((x_reprs, remainder), axis=1)
    return (x_reprs, y_reprs)
  return tl.Fn('Deinterleave', deinterleave, n_out=2)


def RepresentationMask(serializer):
  """Upsamples a mask to cover the serialized representation."""
  # Trax enforces the mask to be of the same size as the target. Get rid of the
  # extra dimensions.
  def representation_mask(mask):
    # mask shape (batch_size,4)
    mask = jnp.amax(mask, axis=tuple(range(2, mask.ndim)))
    # mask shape (batch_size,4)
    mask = jnp.repeat(
        mask[..., jnp.newaxis],
        repeats=serializer.representation_length,
        axis=2)
    # mask shape (batch_size,4,representation_length)
    return mask
  return tl.Fn('RepresentationMask', representation_mask)


def SignificanceWeights(serializer, decay):
  """Multiplies a binary mask with a symbol significance mask."""
  def significance_weights(mask):
    # (repr,) -> (batch, length, repr)
    # significance = [0, 1, 2]
    significance = serializer.significance_map
    assert significance.shape[0] == mask.shape[2]
    # significance = batch_size * [0, 1, 2]
    significance = jnp.repeat(
        significance[np.newaxis, ...], repeats=mask.shape[0], axis=0)
    # significance = batch_size * [0, 1, 2] * mask.shape[1]
    significance = jnp.repeat(
        significance[..., jnp.newaxis], repeats=mask.shape[1], axis=2)
    # significance = batch_size *  mask.shape[1] * [0, 1, 2]
    significance = jnp.swapaxes(significance, 1, 2)
    assert significance.shape == mask.shape
    sig_weights = mask * decay ** significance
    return sig_weights
  return tl.Fn('SignificanceWeights', significance_weights)


class SerializedModel(tl.Serial):
  """Wraps a world model in serialization machinery for training.

  The resulting model takes as input the observation and action sequences,
  serializes them and interleaves into one sequence, which is fed into a given
  autoregressive model. The resulting logit sequence is deinterleaved into
  observations and actions, and the observation logits are returned together
  with computed symbol significance weights.

  The model has a signature
  (obs, act, obs, mask) -> (obs_logits, obs_repr, weights), where obs are
  observations (the second occurrence is the target), act are actions, mask is
  the observation mask, obs_logits are logits of the output observation
  representation, obs_repr is the target observation representation and weights
  are the target weights.
  """

  def __init__(
      self,
      seq_model,
      observation_serializer,
      action_serializer,
      significance_decay,
      mode='train',
  ):
    """Initializes SerializedModel.

    Args:
      seq_model: Trax autoregressive model taking as input a sequence of symbols
        and outputting a sequence of symbol logits.
      observation_serializer: Serializer to use for observations.
      action_serializer: Serializer to use for actions.
      significance_decay: Float from (0, 1) for exponential weighting of symbols
        in the representation.
      mode: 'train' or 'eval'.
    """
    assert mode in ('train', 'eval')
    weigh_by_significance = [
        # (mask,)
        RepresentationMask(serializer=observation_serializer),
        # (repr_mask)
        SignificanceWeights(serializer=observation_serializer,
                            decay=significance_decay),
        # (mask, sig_weights)
    ]
    super().__init__(
        # (obs, act, obs, mask)
        tl.Parallel(Serialize(serializer=observation_serializer),
                    Serialize(serializer=action_serializer),
                    Serialize(serializer=observation_serializer)),
        # (obs_repr, act_repr, obs_repr, mask)
        Interleave(),
        # (obs_act_repr, obs_repr, mask)
        seq_model(mode=mode),
        # (obs_act_logits, obs_repr, mask)
        Deinterleave(x_size=observation_serializer.representation_length,
                     y_size=action_serializer.representation_length),
        # (obs_logits, act_logits, obs_repr, mask)
        tl.Parallel(None, tl.Drop(), None, weigh_by_significance),
        # (obs_logits, obs_repr, weights)
    )

    self._seq_model = seq_model
    self._observation_serializer = observation_serializer
    self._action_serializer = action_serializer

  @property
  def observation_serializer(self):
    return self._observation_serializer

  @property
  def action_serializer(self):
    return self._action_serializer

  def make_predict_model(self):
    """Returns a predict-mode model of the same architecture."""
    return self._seq_model(mode='predict')

  @property
  def seq_model_weights(self):
    """Extracts the weights of the underlying sequence model."""
    return self.weights[2]

  @property
  def seq_model_state(self):
    """Extracts the state of the underlying sequence model."""
    return self.state[2]


def TimeSeriesModel(
    seq_model,
    low=0.0,
    high=1.0,
    precision=2,
    vocab_size=64,
    significance_decay=0.7,
    mode='train',
):
  """Simplified constructor for SerializedModel, for time series prediction."""
  # Model scalar time series.
  obs_srl = space_serializer.BoxSpaceSerializer(
      space=gym.spaces.Box(shape=(), low=low, high=high),
      vocab_size=vocab_size,
      precision=precision,
  )
  # Artifact of the fact that we must provide some actions.
  # TODO(pkozakowski): Remove this requirement.
  act_srl = space_serializer.DiscreteSpaceSerializer(
      space=gym.spaces.Discrete(n=1), vocab_size=1
  )
  seq_model = functools.partial(seq_model, vocab_size=vocab_size)
  return SerializedModel(seq_model, obs_srl, act_srl, significance_decay, mode)


def RawPolicy(seq_model, n_controls, n_actions):
  """Wraps a sequence model in a policy interface.

  The resulting model takes as input observation anc action sequences, but only
  uses the observations. Adds output heads for action logits and value
  predictions.

  Args:
    seq_model: Trax sequence model taking as input and outputting a sequence of
      continuous vectors.
    n_controls: Number of controls.
    n_actions: Number of action categories in each control.

  Returns:
    A model of signature (obs, act) -> (act_logits, values), with shapes:
      obs: (batch_size, length + 1, obs_depth)
      act: (batch_size, length, n_controls)
      act_logits: (batch_size, length, n_controls, n_actions)
      values: (batch_size, length)
  """

  def SplitControls():  # pylint: disable=invalid-name
    """Splits logits for actions in different controls."""
    def f(x):
      return jnp.reshape(x, x.shape[:2] + (n_controls, n_actions))
    return tl.Fn('SplitControls', f)

  action_head = [
      # Predict all action logits at the same time.
      tl.Dense(n_controls * n_actions),
      # Then group them into separate controls, adding a new dimension.
      SplitControls(),
      tl.LogSoftmax(),
  ]
  return tl.Serial(                             # (obs, act)
      tl.Select([0], n_in=2),                   # (obs,)
      seq_model,                                # (obs_hidden,)
      tl.Dup(),                                 # (obs_hidden, obs_hidden)
      tl.Parallel(action_head, [tl.Dense(1),
                                tl.Flatten()])  # (act_logits, values)
  )


def substitute_inner_policy_raw(raw_policy, inner_policy):  # pylint: disable=invalid-name
  """Substitutes the weights/state of the inner model in a RawPolicy."""
  return raw_policy[:1] + [inner_policy] + raw_policy[2:]


def SerializedPolicy(
    seq_model, n_controls, n_actions, observation_serializer, action_serializer
):
  """Wraps a policy in serialization machinery for training.

  The resulting model takes as input observation and action sequences, and
  serializes them into one sequence similar to SerializedModel, before passing
  to the given sequence model. Adds output heads for action logits and value
  predictions.

  Args:
    seq_model: Trax sequence model taking as input a sequence of symbols and
      outputting a sequence of continuous vectors.
    n_controls: Number of controls.
    n_actions: Number of action categories in each control.
    observation_serializer: Serializer to use for observations.
    action_serializer: Serializer to use for actions.

  Returns:
    A model of signature (obs, act) -> (act_logits, values), same as in
    RawPolicy.
  """
  if action_serializer.representation_length != n_controls:
    raise ValueError(
        'Action symbols should correspond 1-1 to controls, but got {} '
        'controls and {} symbols.'.format(
            n_controls, action_serializer.representation_length
        )
    )

  def FirstSymbol():
    return tl.Fn('FirstSymbol', lambda x: x[:, :, 0])

  def PadRight(n_to_pad):
    def pad_right(x):
      pad_widths = [(0, 0), (0, n_to_pad)] + [(0, 0)] * (x.ndim - 2)
      return jnp.pad(
          x, pad_widths, mode='constant', constant_values=x.dtype.type(0))
    return tl.Fn(f'PadRight({n_to_pad})', pad_right)

  action_head = [
      tl.Dense(n_actions),
      tl.LogSoftmax(),
  ]
  value_head = [
      # Take just the vectors corresponding to the first action symbol.
      FirstSymbol(),
      # Predict values.
      tl.Dense(1),
      # Get rid of the singleton dimension.
      tl.Flatten(),
  ]
  return tl.Serial(
      # (obs, act)
      tl.Parallel(Serialize(observation_serializer),
                  Serialize(action_serializer)),
      # (obs_repr, act_repr)
      Interleave(),
      # (obs_act_repr,)

      # Add one dummy action to the right - we'll use the output at its first
      # symbol to predict the value for the last observation.
      PadRight(action_serializer.representation_length),

      # Shift one symbol to the right, so we predict the n-th action symbol
      # based on action symbols 1..n-1 instead of 1..n.
      tl.ShiftRight(),
      seq_model,
      # (obs_act_hidden,)
      Deinterleave(observation_serializer.representation_length,
                   action_serializer.representation_length),
      # (obs_hidden, act_hidden)
      tl.Select([1, 1]),
      # (act_hidden, act_hidden)
      tl.Parallel(action_head, value_head),
      # (act_logits, values)
  )


def substitute_inner_policy_serialized(serialized_policy, inner_policy):  # pylint: disable=invalid-name
  """Substitutes the weights/state of the inner model in a SerializedPolicy."""
  return serialized_policy[:4] + [inner_policy] + serialized_policy[5:]


def analyze_action_space(action_space):  # pylint: disable=invalid-name
  """Returns the number of controls and actions for an action space."""
  assert isinstance(
      action_space, (gym.spaces.Discrete, gym.spaces.MultiDiscrete)
  ), 'Action space expected to be Discrete of MultiDiscrete, got {}.'.format(
      type(action_space)
  )
  if isinstance(action_space, gym.spaces.Discrete):
    n_actions = action_space.n
    n_controls = 1
  else:
    (n_controls,) = action_space.nvec.shape
    assert n_controls > 0
    assert np.min(action_space.nvec) == np.max(action_space.nvec), (
        'Every control must have the same number of actions.'
    )
    n_actions = action_space.nvec[0]
  return (n_controls, n_actions)


def wrap_policy(seq_model, observation_space, action_space, vocab_size):  # pylint: disable=invalid-name
  """Wraps a sequence model in either RawPolicy or SerializedPolicy.

  Args:
    seq_model: Trax sequence model.
    observation_space: Gym observation space.
    action_space: Gym action space.
    vocab_size: Either the number of symbols for a serialized policy, or None.

  Returns:
    RawPolicy if vocab_size is None, else SerializedPolicy.
  """
  (n_controls, n_actions) = analyze_action_space(action_space)
  if vocab_size is None:
    policy_wrapper = RawPolicy
  else:
    obs_serializer = space_serializer.create(observation_space, vocab_size)
    act_serializer = space_serializer.create(action_space, vocab_size)
    policy_wrapper = functools.partial(SerializedPolicy,
                                       observation_serializer=obs_serializer,
                                       action_serializer=act_serializer)
  return policy_wrapper(seq_model, n_controls, n_actions)


def substitute_inner_policy(wrapped_policy, inner_policy, vocab_size):  # pylint: disable=invalid-name
  """Substitutes the inner weights/state in a {Raw,Serialized}Policy.

  Args:
    wrapped_policy (pytree): Weights or state of a wrapped policy.
    inner_policy (pytree): Weights or state of an inner policy.
    vocab_size (int or None): Vocabulary size of a serialized policy, or None
      in case of a raw policy.

  Returns:
    New weights or state of wrapped_policy, with the inner weights/state
      copied from inner_policy.
  """
  if vocab_size is None:
    substitute_fn = substitute_inner_policy_raw
  else:
    substitute_fn = substitute_inner_policy_serialized
  return substitute_fn(wrapped_policy, inner_policy)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for trax.supervised.history."""

from absl.testing import absltest

from trax.supervised import history as trax_history


class HistoryTest(absltest.TestCase):

  def test_unknown_mode(self):
    history = trax_history.History()
    history.append('train', 'metric1', 1, 0.1)
    self.assertEqual(history.get('unknown_mode', 'metric1'), [])

  def test_unknown_metric(self):
    history = trax_history.History()
    history.append('train', 'metric1', 1, 0.1)
    self.assertEqual(history.get('train', 'unknown_metric'), [])

  def test_serializer_and_deserializer(self):
    history = trax_history.History()
    history.append('train', 'metric1', 1, 0.1)
    json_object = history.to_dict()
    history2 = trax_history.History.from_dict(json_object)
    self.assertEqual(history2.get('train', 'metric1'), [(1, 0.1)])

  def test_modes(self):
    history = trax_history.History()
    history.append('train', 'metric1', 1, 0.1)
    history.append('test', 'metric2', 2, 0.2)
    self.assertEqual(history.modes, ['test', 'train'])

  def test_metrics_for_mode(self):
    history = trax_history.History()
    history.append('train', 'metric1', 1, 0.1)
    history.append('train', 'metric2', 2, 0.2)
    self.assertEqual(history.metrics_for_mode('train'), ['metric1', 'metric2'])


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for trax.supervised.trainer_lib."""

import functools
import os

from absl.testing import absltest
from absl.testing import parameterized
import jax
from jax import test_util  # pylint: disable=unused-import
from jax.config import config
import tensorflow.compat.v2 as tf
from trax import fastmath
from trax import layers as tl
from trax import models
from trax import optimizers as trax_opt
from trax import shapes as trax_shapes
from trax import test_utils
from trax.data import inputs as inputs_lib
from trax.fastmath import numpy as jnp
from trax.supervised import lr_schedules as lr
from trax.supervised import trainer_lib
from trax.tf_numpy import extensions as npe
from trax.tf_numpy import numpy as tf_np



def _test_inputs(n_classes, with_weights=False, input_shape=(6, 6, 3)):
  """Make trainer_lib.inputs.Inputs."""
  batch_size = 2 * jax.device_count()

  def input_stream(n_devices):
    del n_devices
    key = fastmath.random.get_prng(0)
    while True:
      keys = fastmath.random.split(key, 4)
      key = keys[0]
      inputs = fastmath.random.uniform(
          keys[1], [batch_size] + list(input_shape))
      targets = fastmath.random.randint(
          keys[2], [batch_size], dtype=jnp.int32, minval=0, maxval=n_classes)
      weights = fastmath.random.uniform(keys[3], [batch_size])
      if with_weights:
        yield inputs, targets, weights
      else:
        yield inputs, targets

  def input_stream_masked(n_devices):
    return inputs_lib.add_loss_weights(input_stream(n_devices))

  return inputs_lib.Inputs(input_stream_masked)


def _test_inputs_lm(vocab_size, seq_len, per_device_batch_size=2):
  """Make trainer_lib.inputs.Inputs for language model."""
  batch_size = per_device_batch_size * jax.device_count()

  def input_stream(_):
    def make_batch(key):
      return fastmath.random.randint(
          key, [batch_size, seq_len], dtype=jnp.int32, minval=0,
          maxval=vocab_size)
    key = fastmath.random.get_prng(0)
    while True:
      keys = fastmath.random.split(key, 3)
      key = keys[0]
      inputs = make_batch(keys[1])
      targets = make_batch(keys[2])
      yield inputs, targets

  def input_stream_masked(n_devices):
    return inputs_lib.add_loss_weights(input_stream(n_devices))

  return inputs_lib.Inputs(input_stream_masked)



BACKENDS = [fastmath.Backend.JAX]
BACKENDS_AND_CONFIGS = [(fastmath.Backend.JAX, [('Simple', None)])]


def short_name(b):
  if b == fastmath.Backend.JAX:
    return 'jax'
  else:
    return 'tf'


def opt_name(opt):
  if opt is None:
    return 'None'
  return opt.__name__


def _pure_lsh_self_attention_fn(n_chunks_after=0):
  return functools.partial(
      tl.PureLSHSelfAttentionWrapper,
      attention_dropout=0.1,
      chunk_len=16,
      n_buckets=[32, 32],
      n_chunks_after=n_chunks_after,
      n_chunks_before=1,
      n_hashes=2,
      n_parallel_heads=1,
      max_length_for_buckets=1024,
      predict_drop_len=128,
      predict_mem_len=1024,
      num_weights=2,
      bias=False,
      pure_lsh_implementation=tl.PureLSHSelfAttention,
  )


def _mixed_lsh_self_attention_fn(n_chunks_after=0):
  return functools.partial(
      tl.PureLSHSelfAttentionWrapper,
      attention_dropout=0.1,
      chunk_len=16,
      n_buckets=[32, 32],
      n_chunks_after=n_chunks_after,
      n_chunks_before=1,
      n_hashes=2,
      n_parallel_heads=1,
      max_length_for_buckets=1024,
      predict_drop_len=128,
      predict_mem_len=1024,
      num_weights=2,
      bias=False,
      pure_lsh_implementation=tl.MixedLSHSelfAttention,
  )


class TraxTest(parameterized.TestCase):

  def __init__(self, methodName='runTest'):  # pylint: disable=invalid-name
    super().__init__(methodName)
    if npe.tpu_devices():
      # Initialize TPU for TF
      resolver = tf.distribute.cluster_resolver.TPUClusterResolver(tpu='local')
      tf.tpu.experimental.initialize_tpu_system(resolver)

  def setUp(self):
    super().setUp()
    test_utils.ensure_flag('test_tmpdir')
    self._old_is_allow_float64 = tf_np.is_allow_float64()
    tf_np.set_allow_float64(False)

  def tearDown(self):
    tf_np.set_allow_float64(self._old_is_allow_float64)
    super().tearDown()

  def _test_train_eval_predict(self, backend, model_name='Simple',
                               optimizer=None):
    with fastmath.use_backend(backend):
      # Prepare model and inputs
      steps = 2
      eval_steps = 2

      if model_name == 'Simple':
        n_classes = 4
        # Adds Dropout and BatchNorm to test state handling.
        def model_fn(mode='train'):
          return tl.Serial(
              tl.Dropout(mode=mode, rate=0.1),
              tl.BatchNorm(mode=mode),
              models.MLP(layer_widths=(16, 16, n_classes), mode=mode))
        inputs = _test_inputs(n_classes)
        n_in = 1
      elif model_name == 'Resnet50':
        n_classes = 4
        model_fn = models.Resnet50
        inputs = _test_inputs(n_classes, input_shape=(224, 224, 3))
        n_in = 1
      elif model_name == 'Transformer':
        vocab_size = 32
        seq_len = 16
        inputs = _test_inputs_lm(vocab_size, seq_len)
        model_fn = functools.partial(
            models.Transformer,
            input_vocab_size=vocab_size)
        n_in = 2
      else:
        raise ValueError('Unrecognized model name: ' + model_name)

      kwargs = {}
      if optimizer is not None:
        kwargs['optimizer'] = optimizer

      # Train and evaluate
      output_dir = self.create_tempdir().full_path
      loop = trainer_lib.train(
          output_dir,
          model=model_fn,
          inputs=inputs,
          steps=steps,
          eval_steps=eval_steps,
          eval_frequency=1,  # eval at every step.
          **kwargs)

      # Assert total train steps
      self.assertEqual(steps, loop.step)

      inputs = inputs.train_stream(1)

      # Predict with final weights
      model = model_fn()
      weights = loop.model.weights
      state = loop.model.state
      model(next(inputs)[:n_in], weights=weights, state=state)

      # Predict with weights loaded from file.
      model = model_fn()
      model.init_from_file(os.path.join(output_dir, 'model.pkl.gz'))
      model(next(inputs)[:n_in])

  @parameterized.named_parameters(
      ('_%s_%s_%s' % (short_name(backend), model_name, opt_name(opt)),  # pylint: disable=g-complex-comprehension
       backend, model_name, opt)
      for backend, configs in BACKENDS_AND_CONFIGS
      for model_name, opt in configs)
  def test_train_eval_predict(self, backend, model_name, opt):
    self._test_train_eval_predict(backend, model_name, opt)

  @parameterized.parameters(BACKENDS)
  def test_train_eval_predict_sm3(self, backend):
    self._test_train_eval_predict(backend, 'Simple', trax_opt.SM3)

  @parameterized.parameters(BACKENDS)
  def test_train_restart(self, backend):
    with fastmath.use_backend(backend):
      # Prepare model and inputs
      n_classes = 4
      steps = 2
      eval_steps = 2
      model_fn = functools.partial(models.MLP,
                                   layer_widths=(16, 16, n_classes))
      inputs = _test_inputs(n_classes)

      # Train and evaluate
      output_dir = self.create_tempdir().full_path
      trainer_lib.train(
          output_dir,
          model=model_fn,
          inputs=inputs,
          steps=steps,
          eval_steps=eval_steps,
          eval_frequency=1)

      # Restart training
      loop = trainer_lib.train(
          output_dir,
          model=model_fn,
          inputs=inputs,
          steps=(2 * steps),
          eval_steps=eval_steps,
          eval_frequency=1)

      # Assert total train steps
      self.assertEqual(loop.step, 2 * steps)

  @parameterized.parameters(BACKENDS)
  def test_train_permanent_checkpoints(self, backend):
    with fastmath.use_backend(backend):
      # Prepare model and inputs
      n_classes = 4
      steps = 5
      eval_steps = 2
      model_fn = functools.partial(models.MLP,
                                   layer_widths=(16, 16, n_classes))
      inputs = _test_inputs(n_classes)

      # Train and evaluate
      output_dir = self.create_tempdir().full_path

      # Steps 1 -> 5
      loop = trainer_lib.train(
          output_dir,
          model=model_fn,
          inputs=inputs,
          steps=steps,
          eval_steps=eval_steps,
          eval_frequency=1,
          permanent_checkpoint_frequency=2)

      # Steps 6 -> 10
      loop = trainer_lib.train(
          output_dir,
          model=model_fn,
          inputs=inputs,
          steps=(2 * steps),
          eval_steps=eval_steps,
          eval_frequency=1,
          permanent_checkpoints_at=[7, 8, 10])

      path = os.path.join(output_dir, 'model.pkl.gz')
      self.assertTrue(tf.io.gfile.exists(path))

      for step in range(11):
        filename = 'model_{}.pkl.gz'.format(step)
        path = os.path.join(output_dir, filename)
        if step in [1, 2, 4, 7, 8, 10]:
          self.assertTrue(tf.io.gfile.exists(path),
                          msg='No model for step: {} in dir {}.'.format(
                              step, tf.io.gfile.listdir(output_dir)))
        else:
          self.assertFalse(tf.io.gfile.exists(path),
                           msg='Model for step: {} in dir {}.'.format(
                               step, tf.io.gfile.listdir(output_dir)))

      # Assert total train steps
      self.assertEqual(loop.step, 10)

  @parameterized.parameters(BACKENDS)
  def test_train_restart_with_same_steps(self, backend):
    with fastmath.use_backend(backend):
      # Prepare model and inputs
      n_classes = 4
      steps = 2
      eval_steps = 2
      model_fn = functools.partial(models.MLP,
                                   layer_widths=(16, 16, n_classes))
      inputs = _test_inputs(n_classes)

      # Train and evaluate
      output_dir = self.create_tempdir().full_path
      trainer_lib.train(
          output_dir,
          model=model_fn,
          inputs=inputs,
          steps=steps,
          eval_steps=eval_steps,
          eval_frequency=1)

      # Restart training
      loop = trainer_lib.train(
          output_dir,
          model=model_fn,
          inputs=inputs,
          steps=steps,
          eval_steps=eval_steps,
          eval_frequency=1)

      # Assert total train steps
      self.assertEqual(loop.step, steps)

  def test_train_with_pure_lsh_attention(self, backend=fastmath.Backend.JAX):
    with fastmath.use_backend(backend):
      # Prepare model and inputs
      def model(mode='train'):
        return models.ConfigurableTerraformer(
            mode=mode,
            d_model=16,
            d_ff=16,
            n_heads=2,
            dropout=0.05,
            n_decoder_layers=1,
            n_encoder_layers=1,
            input_vocab_size=256,
            encoder_attention_type=_pure_lsh_self_attention_fn(),
            encoder_decoder_attention_type=_pure_lsh_self_attention_fn(),
        )

      max_len = 128
      inputs = _test_inputs_lm(vocab_size=256, seq_len=max_len)

      steps = 1
      eval_steps = 1

      # Train and evaluate
      output_dir = self.create_tempdir().full_path
      trainer_lib.train(
          output_dir,
          model=model,
          inputs=inputs,
          steps=steps,
          eval_steps=eval_steps,
          eval_frequency=1)

      # Read checkpoint
      model_file = os.path.join(output_dir, 'model.pkl.gz')

      shape11 = trax_shapes.ShapeDtype((1, 1), dtype=jnp.int32)
      shape1l = trax_shapes.ShapeDtype((1, max_len), dtype=jnp.int32)

      model_predict = model(mode='predict')
      model_predict.init_from_file(
          model_file, weights_only=True, input_signature=(shape1l, shape11))

  def test_train_with_mixed_lsh_attention(self, backend=fastmath.Backend.JAX):
    with fastmath.use_backend(backend):
      # Prepare model and inputs

      def model(mode='train'):
        return models.ConfigurableTerraformer(
            mode=mode,
            d_model=16,
            d_ff=16,
            n_heads=2,
            dropout=0.05,
            n_decoder_layers=1,
            n_encoder_layers=1,
            input_vocab_size=256,
            encoder_attention_type=_mixed_lsh_self_attention_fn(),
            encoder_decoder_attention_type=_mixed_lsh_self_attention_fn(),
        )

      max_len = 128
      inputs = _test_inputs_lm(vocab_size=256, seq_len=max_len)

      steps = 1
      eval_steps = 1

      # Train and evaluate
      output_dir = self.create_tempdir().full_path
      trainer_lib.train(
          output_dir,
          model=model,
          inputs=inputs,
          steps=steps,
          eval_steps=eval_steps,
          eval_frequency=1)

      # Read checkpoint
      model_file = os.path.join(output_dir, 'model.pkl.gz')

      shape11 = trax_shapes.ShapeDtype((1, 1), dtype=jnp.int32)
      shape1l = trax_shapes.ShapeDtype((1, max_len), dtype=jnp.int32)

      model_predict = model(mode='predict')
      model_predict.init_from_file(model_file, weights_only=True,
                                   input_signature=(shape1l, shape11))

  @parameterized.parameters(BACKENDS)
  def test_train_fills_in_missing_eval_metrics(self, backend):
    with fastmath.use_backend(backend):
      # Prepare model and inputs
      n_classes = 4
      steps = 2
      eval_steps = 2
      model_fn = functools.partial(models.MLP, layer_widths=(16, 16, n_classes))
      inputs = _test_inputs(n_classes)
      additional_eval_stream = trainer_lib.NamedStream(
          # deliberately duplicating eval data
          stream=inputs.eval_stream(1),
          name='additional_eval_task')

      # Train and evaluate
      output_dir = self.create_tempdir().full_path
      loop = trainer_lib.train(
          output_dir,
          model=model_fn,
          inputs=inputs,
          steps=steps,
          eval_steps=eval_steps,
          eval_frequency=1,
          additional_eval_streams=[additional_eval_stream])

      self.assertLen(loop.eval_tasks, 2)
      eval_task_1, eval_task_2 = loop.eval_tasks
      self.assertCountEqual(eval_task_1.metrics, eval_task_2.metrics)
      self.assertCountEqual(eval_task_1.metric_names, eval_task_2.metric_names)

  @parameterized.named_parameters(
      ('_%s' % short_name(backend), backend)
      for backend in BACKENDS)
  def test_train_with_weights(self, backend):
    with fastmath.use_backend(backend):
      # Prepare model and inputs
      n_classes = 4
      steps = 2
      eval_steps = 2
      model_fn = functools.partial(models.MLP,
                                   layer_widths=(16, 16, n_classes))
      inputs = _test_inputs(n_classes, with_weights=True)

      # Train and evaluate
      output_dir = self.create_tempdir().full_path
      state = trainer_lib.train(
          output_dir,
          model=model_fn,
          inputs=inputs,
          steps=steps,
          eval_steps=eval_steps)

      # Assert total train steps
      self.assertEqual(state.step, steps)

  @parameterized.parameters(BACKENDS)
  def test_reset_twice(self, backend):
    with fastmath.use_backend(backend):
      n_classes = 4
      model_fn = functools.partial(models.MLP,
                                   layer_widths=(16, 16, n_classes))
      inputs = _test_inputs(n_classes)

      trainer = trainer_lib.Trainer(
          model=model_fn,
          loss_fn=tl.WeightedCategoryCrossEntropy(),
          optimizer=trax_opt.SM3,
          lr_schedule=lr.multifactor(),
          inputs=inputs,
      )

      output_dir1 = self.create_tempdir(name='output_dir1').full_path
      trainer.reset(output_dir1)
      trainer.evaluate(1)
      output_dir2 = self.create_tempdir(name='output_dir2').full_path
      trainer.reset(output_dir2)
      trainer.evaluate(1)

  def test_tf_xla_forced_compile(self):
    # TODO(wangpeng): re-enable this test
    self.skipTest('Needs --config=cuda to pass this test')
    old_flag = fastmath.tf.tf_xla_forced_compile_enabled()
    fastmath.tf.set_tf_xla_forced_compile(True)
    self._test_train_eval_predict('tf')
    fastmath.tf.set_tf_xla_forced_compile(old_flag)



class EpochsTest(absltest.TestCase):

  def test_cuts_epoch_when_total_steps_reached(self):
    epoch_steps = trainer_lib.epochs(
        total_steps=5, steps_to_skip=0, epoch_steps=[1, 2, 3])
    self.assertEqual(list(epoch_steps), [1, 2, 2])

  def test_skips_full_epoch(self):
    epoch_steps = trainer_lib.epochs(
        total_steps=4, steps_to_skip=2, epoch_steps=[2, 2])
    self.assertEqual(list(epoch_steps), [2])

  def test_skips_part_of_epoch(self):
    epoch_steps = trainer_lib.epochs(
        total_steps=4, steps_to_skip=1, epoch_steps=[2, 2])
    self.assertEqual(list(epoch_steps), [1, 2])


if __name__ == '__main__':
  config.config_with_absl()
  tf.compat.v1.enable_eager_execution()
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for supervised training optimizers."""

from absl.testing import absltest

import numpy as np

from trax import optimizers
from trax.optimizers import momentum


class OptimizersTest(absltest.TestCase):

  def test_slots(self):
    weights_shape = (3, 5)
    weight_tree = np.arange(15).reshape(weights_shape)

    # SGD - an optimizer that doesn't use slots.
    opt_1 = optimizers.SGD(.01)
    self.assertIsNone(opt_1.slots)
    opt_1.tree_init(weight_tree)
    self.assertIsInstance(opt_1.slots, tuple)
    self.assertLen(opt_1.slots, 1)
    self.assertIsNone(opt_1.slots[0])

    # Momentum - an optimizer with slots
    opt_2 = momentum.Momentum(.01)
    self.assertIsNone(opt_2.slots)
    opt_2.tree_init(weight_tree)
    self.assertIsInstance(opt_2.slots, tuple)
    self.assertLen(opt_2.slots, 1)
    self.assertEqual(weights_shape, opt_2.slots[0].shape)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Random functions."""
import numpy as np
import tensorflow.compat.v2 as tf

from trax.tf_numpy.numpy_impl import utils


DEFAULT_RANDN_DTYPE = np.float32


def randn(*args):
  """Returns samples from a normal distribution.

  Uses `tf.random_normal`.

  Args:
    *args: The shape of the output array.

  Returns:
    An ndarray with shape `args` and dtype `float64`.
  """
  # TODO(wangpeng): Use new stateful RNG
  if utils.isscalar(args):
    args = (args,)
  return utils.tensor_to_ndarray(
      tf.random.normal(args, dtype=DEFAULT_RANDN_DTYPE))


def seed(s):
  """Sets the seed for the random number generator.

  Uses `tf.set_random_seed`.

  Args:
    s: an integer.
  """
  # TODO(wangpeng): make the signature the same as numpy
  tf.random.set_seed(s)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for trax.rl.normalization."""

from absl.testing import absltest
import numpy as np

from trax import shapes
from trax.rl import normalization


class NormalizationTest(absltest.TestCase):

  def test_running_mean(self):
    x = np.random.uniform(size=10)
    state = normalization.running_mean_init(shape=())
    for i in range(len(x)):
      state = normalization.running_mean_update(x[i], state)
      np.testing.assert_almost_equal(
          normalization.running_mean_get_mean(state), np.mean(x[:i + 1])
      )

  def test_running_variance(self):
    x = np.random.uniform(size=10)
    state = normalization.running_mean_and_variance_init(shape=())
    for i in range(len(x)):
      state = normalization.running_mean_and_variance_update(x[i], state)
      np.testing.assert_almost_equal(
          normalization.running_mean_and_variance_get_variance(state),
          np.var(x[:i + 1]),
      )

  def test_normalize_collect(self):
    x = np.random.uniform(size=(2, 3, 4, 5))
    normalize = normalization.Normalize(mode='collect')
    normalize.init(shapes.signature(x))
    old_state = normalize.state
    y = normalize(x)
    with self.assertRaises(AssertionError):
      np.testing.assert_equal(normalize.state, old_state)
    with self.assertRaises(AssertionError):
      np.testing.assert_almost_equal(x, y)

  def test_normalize_train(self):
    x = np.random.uniform(size=(2, 3, 4, 5))
    normalize = normalization.Normalize(mode='train', epsilon=0.0)
    normalize.init(shapes.signature(x))
    old_state = normalize.state
    y = normalize(x)
    np.testing.assert_equal(normalize.state, old_state)
    np.testing.assert_almost_equal(x, y)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Original API for supervised learning/training in Trax.

Trax authors expect that the `supervised.training` module (under development)
will replace `trainer_lib`.
"""

import collections
import functools
import itertools
import os
import sys
import time

from absl import logging

import gin

import jax
import tensorflow.compat.v2 as tf
from trax import fastmath
from trax import jaxboard
from trax import layers as tl
from trax import optimizers as trax_opt
from trax.data import inputs as trax_inputs
from trax.fastmath import numpy as np
from trax.fastmath import random as jax_random
from trax.layers import base
from trax.shapes import ShapeDtype
from trax.supervised import history as trax_history
from trax.supervised import lr_schedules as lr
from trax.supervised import training


# TODO(afrozm): Maybe flatten everything from OptState into TrainerState.
TrainerState = collections.namedtuple('_TrainerState', [
    'step',         # Current training step number.
    'opt_state',    # OptState.
    'history',      # trax.history.History.
    'model_state',  # Auxilliary state of the model.
])


OptState = collections.namedtuple('_OptState', [
    'weights',     # Model weights.
    'slots',       # Per-parameter optimizer state, e.g. gradient moments.
    'opt_params',  # Optimizer (hyper)parameters, e.g. learning rate, momentum.
])


_DEFAULT_METRICS = {
    'loss': tl.WeightedCategoryCrossEntropy(),
    'accuracy': tl.WeightedCategoryAccuracy(),
    'sequence_accuracy': tl.MaskedSequenceAccuracy(),
    'neg_log_perplexity': tl.Serial(tl.WeightedCategoryCrossEntropy(),
                                    tl.Negate()),
    'weights_per_batch_per_core': tl.Serial(tl.Drop(), tl.Drop(), tl.Sum()),
}


NamedStream = collections.namedtuple(
    'NamedStream', ['name', 'stream']
)


@gin.configurable
def named_stream(name=gin.REQUIRED, stream=gin.REQUIRED):
  return NamedStream(name=name, stream=stream)


class Trainer:
  """Trax trainer.

  A trainer allows to make training steps, train for full epochs,
  save the training state and access evaluation data.
  """

  def __init__(self, model, loss_fn, optimizer, lr_schedule, inputs,
               output_dir=None, random_seed=None, n_devices=None,
               checkpoints_at=None, should_save_checkpoints=True,
               should_write_summaries=True,
               metrics=None, checkpoint_highest=None,
               checkpoint_lowest=None,
               init_checkpoint=None):

    self._is_chief, _, self._n_devices, rng = (
        training.init_host_and_devices(n_devices, random_seed))
    self._should_save_checkpoints = should_save_checkpoints and self._is_chief
    self._checkpoints_at = checkpoints_at if checkpoints_at is not None else []
    self._should_write_summaries = should_write_summaries
    if not output_dir:
      self._should_save_checkpoints = False
      self._should_write_summaries = False
    self._checkpoint_highest = checkpoint_highest
    self._checkpoint_lowest = checkpoint_lowest
    self._metrics_dict = metrics if metrics is not None else _DEFAULT_METRICS
    # Inputs is either an Inputs instance or a function that returns it.
    self._inputs = inputs
    if callable(inputs):  # If we pass a function, e.g., through gin, call it.
      self._inputs = inputs()
    # Initialize the learning rate to a dummy value. It will be set in reset().
    opt = optimizer(learning_rate=0.0)

    # Setup the model.
    model_train = model(mode='train')
    model_predict_eval = model(mode='eval')
    # Should work for fine-tuning of T5.
    if init_checkpoint:
      model_train.init_from_file(init_checkpoint, weights_only=True)
      model_predict_eval.init_from_file(init_checkpoint, weights_only=True)
    self._model_with_loss = tl.Serial(model_train, loss_fn)

    # Setup state.
    rng, init_rng = jax_random.split(rng)
    self._rngs = np.stack(jax_random.split(rng, self._n_devices))
    shapes, dtypes = self._inputs.example_shape_dtype
    input_signature = tuple(ShapeDtype(s, d) for (s, d) in zip(shapes, dtypes))

    def new_opt_state_and_model_state(rng):
      """Returns optimizer and model states suitable for training a model."""
      weights, state = self._model_with_loss.init(input_signature, rng=rng)
      (slots, opt_params) = opt.tree_init(weights)
      return (OptState(weights, slots, opt_params), state)

    if fastmath.is_backend(fastmath.Backend.JAX):
      # JIT parameter initialization to avoid memory fragmentation
      new_opt_state_and_model_state = (
          fastmath.jit(new_opt_state_and_model_state))
    self._new_opt_state_and_model_state = (
        lambda: new_opt_state_and_model_state(init_rng))

    # Arrange and initialize metrics layers.
    self._metrics = list(sorted(self._metrics_dict.keys()))
    metrics_layers = [self._metrics_dict[m] for m in self._metrics]
    metrics_in_parallel = tl.Branch(*metrics_layers)
    metrics_in_parallel.rng = init_rng
    example_signature = tuple(
        ShapeDtype(s, d) for (s, d) in zip(*self._inputs.example_shape_dtype)
    )
    model_predict_eval.init(example_signature)
    self._input_signature = example_signature
    output_signature = model_predict_eval.output_signature(example_signature)
    m_weights, m_state = metrics_in_parallel.init(output_signature)
    self._metrics_weights = self._for_n_devices(m_weights)
    self._metrics_state = self._for_n_devices(m_state)

    # Jit model_predict and update so they're fast.
    self._jit_eval = _jit_predict_fn(
        model_predict_eval, metrics_in_parallel, self._n_devices)
    self._jit_update_fn = _jit_update_fn(
        model_train, loss_fn, opt, self._n_devices)

    self._model_train = model_train
    self._model_predict_eval = model_predict_eval
    self._loss_fn = loss_fn
    self._lr_schedule = lr_schedule

    # Those fields will be set in reset().
    self._output_dir = None
    self._train_sw = None
    self._eval_sw = None
    self._history = None
    self._opt_state = None
    self._step = None
    self._model_state = None
    self.reset(output_dir)

  @property
  def n_devices(self):
    return self._n_devices

  @property
  def step(self):
    return self._step

  @property
  def model_weights(self):
    # Currently we need to pick [0] as we ignore loss weights (empty).
    weights = self._opt_state.weights[0]
    if self.n_devices > 1:
      unreplicate = lambda x: x[0]
      weights = fastmath.nested_map(unreplicate, weights)
    return weights

  @model_weights.setter
  def model_weights(self, weights):
    new_model_weights = self._for_n_devices(weights)
    if isinstance(self._opt_state.weights, list):
      self._opt_state.weights[0] = new_model_weights
    else:  # weights are a tuple, need to re-create
      new_weights = [new_model_weights] + list(self._opt_state.weights[1:])
      self._opt_state = self._opt_state._replace(weights=new_weights)

  @property
  def model_state(self):
    # Currently we need to pick [0] as we ignore loss state (empty).
    state = self._model_state[0]
    if self.n_devices > 1:
      unreplicate = lambda x: x[0]
      state = fastmath.nested_map(unreplicate, state)
    return state

  @model_state.setter
  def model_state(self, state):
    new_model_state = self._for_n_devices(state)
    if isinstance(self._model_state, list):
      self._model_state[0] = new_model_state
    else:  # weights are a tuple, need to re-create
      self._model_state = [new_model_state] + list(self._model_state[1:])

  @property
  def state(self):
    return TrainerState(
        opt_state=self._opt_state, step=self._step, history=self._history,
        model_state=self._model_state)

  @property
  def learning_rate(self):
    with fastmath.use_backend(fastmath.Backend.NUMPY):
      return self._lr_schedule(self._step)

  def reset(self, output_dir, init_checkpoint=None):
    """Reset the model parameters.

    Restores the parameters from the given output_dir if a checkpoint exists,
    otherwise randomly initializes them.

    Does not re-jit the model.

    Args:
      output_dir: Output directory.
      init_checkpoint: Initial checkpoint (default $output_dir/model.pkl.gz)
    """
    self.close()
    self._output_dir = output_dir
    if output_dir is not None:
      tf.io.gfile.makedirs(output_dir)
    else:
      assert not self._should_save_checkpoints
      assert not self._should_write_summaries

    # Create summary writers and history.
    if self._should_write_summaries:
      self._train_sw = jaxboard.SummaryWriter(os.path.join(output_dir, 'train'),
                                              enable=self._is_chief)
      self._eval_sw = jaxboard.SummaryWriter(os.path.join(output_dir, 'eval'),
                                             enable=self._is_chief)

    # Reset the train and eval streams.
    self._train_stream = _repeat_stream(self._inputs.train_stream,
                                        self._n_devices)
    # TODO(lukaszkaiser): add an option to evaluate exactly on the full eval
    #   set by adding a padding and stopping the stream when too large.
    self._eval_stream = _repeat_stream(
        self._inputs.eval_stream, self._n_devices)
    self._train_eval_stream = _repeat_stream(
        self._inputs.train_eval_stream, self._n_devices)

    # Restore the training state.
    if output_dir is not None:
      state = load_trainer_state(output_dir, self._model_with_loss,
                                 init_checkpoint)
    else:
      state = TrainerState(step=None, opt_state=None,
                           history=trax_history.History(), model_state=None)
    self._step = state.step or 0
    history = state.history
    self._history = history
    if state.opt_state:
      opt_state = state.opt_state
      model_state = state.model_state
    else:
      opt_state, model_state = self._new_opt_state_and_model_state()
      model_state = self._for_n_devices(model_state)
    self._opt_state = OptState(*self._for_n_devices(opt_state))
    self._model_state = model_state
    if not state.opt_state and self._should_save_checkpoints:
      self.save_state(keep=False)

  def train_epoch(self, n_steps, n_eval_steps):
    """Runs `n_steps` of training, with periodic logging, saving, and evals."""
    # TODO(jonni): Clarify how this method relates to the stricter notion of
    # epoch (training for as many steps as needed for a full pass through the
    # training data).
    print()  # Add visual separator in logs for start of training epoch.
    start_time = time.time()

    for _ in range(n_steps):
      batch = next(self._train_stream)
      if self.n_devices > 1:  # TODO(lukaszkaiser): use everywhere if possible.
        batch = _reshape_by_device(batch, self.n_devices)
      self.train_step(batch)
      if self._should_save_now():
        self.save_state(keep=True)
      if self._should_log_now():
        self._train_sw.scalar('training/learning_rate', self.learning_rate)

    # At end of n_steps, do bookkeeping, run evals, and save state.
    elapsed_time = time.time() - start_time
    self.log_step('Ran %d train steps in %0.2f secs' % (n_steps, elapsed_time))
    if self._train_sw and n_steps > 1:
      self._train_sw.scalar('training/steps per second',
                            n_steps / elapsed_time, step=self._step)
      self._train_sw.flush()
    self.evaluate(n_eval_steps)
    if self._eval_sw:
      self._eval_sw.flush()
    if self._should_save_checkpoints:
      self.save_state(keep=False)
    if self._should_save_checkpoints and self._current_step_is_best(high=True):
      self.save_state(keep=False, prefix='highest_' + self._checkpoint_highest)
    if self._should_save_checkpoints and self._current_step_is_best(high=False):
      self.save_state(keep=False, prefix='lowest_' + self._checkpoint_lowest)

  def train_step(self, batch):
    """Run one training step and update self._opt_state."""
    # Calculate the current optimizer parameters.
    opt_param_updates = self._for_n_devices(
        {'learning_rate': np.array(self.learning_rate)})
    opt_state = self._opt_state
    opt_state.opt_params.update(opt_param_updates)

    # Run the update.
    weights, slots, opt_params = opt_state
    (weights, slots), stat, self._model_state, self._rngs = self._jit_update_fn(
        (weights, slots), self._step, opt_params, batch,
        self._model_state, self._rngs)
    self._opt_state = opt_state._replace(weights=weights, slots=slots)
    if self._should_log_now():
      for name, value in stat.items():
        # TODO(afrozm): value is a scalar, but sometimes JAX is crashing here
        # with a device put array error complaining that it should be an array.
        # On multiple devices, take the mean.
        scalar_value = np.mean(np.array(value))
        self._train_sw.scalar('training/' + name, scalar_value, step=self._step)
    self._step += 1

  def evaluate(self, n_eval_steps):
    """Evaluate the model and log metrics."""
    _, rng = jax_random.split(self._rngs[0])
    # TODO(lukaszkaiser): both model state and parameters by default include
    # the loss layer. Currently, we access the pure-model parameters by just
    # indexing, [0] here. But we should make it more explicit in a better API.
    weights = (self._opt_state.weights[0], self._metrics_weights)
    state = (self._model_state[0], self._metrics_state)
    self.log_step('Evaluation')
    train_eval_slice = itertools.islice(self._train_eval_stream, n_eval_steps)
    train_metrics, _ = self.evaluation_round(train_eval_slice, weights, state,
                                             rng)
    self.log_metrics(train_metrics, self._train_sw, 'train')
    eval_slice = itertools.islice(self._eval_stream, n_eval_steps)
    eval_metrics, _ = self.evaluation_round(eval_slice, weights, state, rng)
    self.log_metrics(eval_metrics, self._eval_sw, 'eval')
    self.log_step('Finished evaluation')

    # Save the learning rate in history.
    self._history.append('train', 'training/learning_rate',
                         self._step, self.learning_rate)

  def evaluation_round(self, inputs_stream, weights, state, rng):
    """Evaluate.

    Args:
      inputs_stream: Iterable of inputs to evaluate on.
      weights: Weights for each f in eval_fns.
      state: State for each f in eval_fns.
      rng: Single-use random number generator (JAX PRNG key).

    Returns:
      Tuple of `(metrics, state)`. `metrics` is a dict from metric name to
      metric value averaged over the number of inputs, and `state` is the end
      state returned by this trainer's `predict_fn`.
    """
    metrics = collections.defaultdict(float)
    count = 0
    for inp in inputs_stream:
      count += 1
      rng, subrng = jax_random.split(rng)
      metric_values, _ = self._jit_eval(inp, weights, state, subrng)
      try:
        metric_values = list(metric_values)
      except (TypeError, IndexError):
        metric_values = [float(metric_values)]
      for m, v in zip(self._metrics, metric_values):
        metrics[m] += v
    return {m: v / count for (m, v) in metrics.items()}, state

  def save_gin(self):
    """"Saves the operative gin config, only if it is the chief."""
    if not self._is_chief:
      return
    assert self._output_dir is not None
    config_path = os.path.join(self._output_dir, 'config.gin')
    config_str = gin.operative_config_str()
    with tf.io.gfile.GFile(config_path, 'w') as f:
      f.write(config_str)
    sw = self._train_sw
    if sw:
      sw.text('gin_config',
              jaxboard.markdownify_operative_config_str(config_str))

  def _save_state_dict(self, trainer_state_dict, weights_file):
    training.pickle_to_file(trainer_state_dict, weights_file, gzip=True)
    log('Model saved to %s' % weights_file, stdout=False)

  def save_state(self, keep, prefix='model'):
    """Save trainer state given a possibly replicated opt_state."""
    opt_state = self._opt_state
    if self.n_devices > 1:
      first_replica = lambda x: x[0]
      opt_state = OptState(*fastmath.nested_map(first_replica, opt_state))
    # This line, while optional, allows JAX to transfer arrays from the device
    # to the host in parallel, which is particularly important for cloud TPU.
    if fastmath.is_backend(fastmath.Backend.JAX):
      opt_state = jax.device_get(opt_state)
    step, history, model_state = self._step, self._history, self._model_state
    output_dir = self._output_dir

    weights_file = os.path.join(output_dir, prefix + '.pkl.gz')

    # This dict will be stored as the model.
    trainer_state_dict = make_trainer_state_dict(
        step, opt_state, history, model_state, self._input_signature)
    self._save_state_dict(trainer_state_dict, weights_file)

    if keep:
      weights_file = os.path.join(output_dir,
                                  '{}_{}.pkl.gz'.format(prefix, step))
      self._save_state_dict(trainer_state_dict, weights_file)

  def save_computation_graphs(self):
    """Dump computation graphs to files."""
    if self.n_devices != 1:
      return  # TODO(lukaszkaiser): make this work with more devices.
    batch = next(self._train_stream)
    output_dir = self._output_dir
    if self.n_devices > 1:
      batch = _reshape_by_device(batch, self.n_devices)
    weights = self._opt_state.weights[0]
    forward_computation = jax.xla_computation(self._model_predict_eval)(
        batch, weights=weights, state=self._model_state[0],
        rng=self._rngs[0])
    with tf.io.gfile.GFile(os.path.join(output_dir, 'forward.txt'), 'w') as f:
      f.write(forward_computation.as_hlo_text())
    with tf.io.gfile.GFile(os.path.join(output_dir, 'forward.dot'), 'w') as f:
      f.write(forward_computation.as_hlo_dot_graph())

  def log_step(self, step_message):
    log('Step % 6d: %s' % (self.step, step_message))

  def log_metrics(self, metrics, summ_writer, log_prefix):
    """Log metrics to summary writer and history."""
    history = self._history
    rjust_len = max([0] + [len(name) for name in metrics])
    for name, value in metrics.items():
      self.log_step('%s %s | % .8f' % (
          log_prefix.ljust(5), name.rjust(rjust_len), value))
      full_name = 'metrics/' + name
      if history:
        history.append(log_prefix, full_name, self.step, value)
      if summ_writer:
        summ_writer.scalar(full_name, value, self.step)

  def print_n_weights(self):
    """Prints the total count of trainable weights."""
    opt_state = self._opt_state
    sizes = _sizes(opt_state.weights)
    if self.n_devices > 1:
      unreplicate = lambda x: x[0]
      single_weights = fastmath.nested_map(unreplicate, opt_state.weights)
      sizes = _sizes(single_weights)
    total_size = _nested_reduce(sum, sizes)
    self.log_step('Total number of trainable weights: %d' % total_size)

  def _should_save_now(self):
    return self._should_save_checkpoints and self._step in self._checkpoints_at

  def _current_step_is_best(self, high):
    """Is the current step the best (highest if high, else lowest)."""
    metric = self._checkpoint_highest if high else self._checkpoint_lowest
    if metric is None:
      return False
    # History is a list of pairs (step, value).
    history = self._history.get('eval', 'metrics/' + metric)
    sequence = [float(i[1]) for i in history]  # Just the values.
    best = max(sequence) if high else min(sequence)  # Best value.
    last_is_best = float(history[-1][1]) == best  # Is last the best?
    cur_step = history[-1][0] == self._step  # Is last the current step?
    return cur_step and last_is_best

  def _should_log_now(self):
    return (self._train_sw is not None
            and (self._step == 1 or self._step % 10 == 0))

  def _for_n_devices(self, x):
    """Replicates/broadcasts `x` for n devices if `self.n_devices > 1`."""
    return tl.for_n_devices(x, self.n_devices)  # pylint: disable=protected-access

  def close(self):
    if self._train_sw is not None:
      self._train_sw.close()
      self._train_sw = None
    if self._eval_sw is not None:
      self._eval_sw.close()
      self._eval_sw = None


@gin.configurable(denylist=['output_dir'])
def train(output_dir,
          model=gin.REQUIRED,
          loss_fn=tl.WeightedCategoryCrossEntropy(),
          inputs=trax_inputs.batcher,
          optimizer=trax_opt.Adafactor,
          lr_schedule_fn=lr.multifactor,
          trainer_class=Trainer,
          steps=1000,
          checkpoints_at=None,
          permanent_checkpoints_at=None,
          eval_steps=10,
          eval_frequency=100,
          permanent_checkpoint_frequency=None,
          random_seed=None,
          save_graphs=True,
          metrics=None,
          checkpoint_highest=None,
          checkpoint_lowest=None,
          use_loop=True,
          loss_chunk_size=0,
          use_memory_efficient_trainer=False,
          adasum=False,
          init_checkpoint=None,
          callbacks=None,
          n_weights_shards=1,
          additional_train_tasks=None,
          additional_eval_tasks=None,
          additional_eval_streams=None):
  """Train the model on the inputs.

  Args:
    output_dir: Directory where to put the logs and checkpoints.
    model: The model to train as a callable returning 2 callables, an init_fn
      and apply_fn.
    loss_fn: callable with signature: weights, trax.inputs.Inputs, model, state,
      rng -> loss.
    inputs: callable returning trax.inputs.Inputs.
    optimizer: The optimizer (see optimizers/base.py for signature).
    lr_schedule_fn: A learning rate schedule function, that when called returns
      a function from step to learning rate (a float).
    trainer_class: The trainer class to use.
    steps: int, total number of training steps.
    checkpoints_at: list of integers. Save a checkpoint for each training step
      in the list.
    permanent_checkpoints_at: list of integers. Save a permanent checkpoint for
      each training step in the list.
    eval_steps: int, num of steps per evaluation. If None or 0, eval disabled.
    eval_frequency: int, how often to run evaluation (every eval_frequency
      steps). If None or 0, eval disabled.
    permanent_checkpoint_frequency: int, how often to save permanent checkpoints
      (every permanent_checkpoint_frequency steps).
    random_seed: the random seed to use; time/os dependent if None (default).
    save_graphs: bool, if True, save computation graph to file.
    metrics: optionally override the default metrics dictionary.
    checkpoint_highest: save the checkpoint highest at this metric.
    checkpoint_lowest: save the checkpoint lowest at this metric.
    use_loop: whether to use training.Loop instead of Trainer.
    loss_chunk_size: int, if > 0 chunk loss into these sizes to save memory.
    use_memory_efficient_trainer: whether to use memory-efficient trainer.
    adasum: if True, use adaptive summation for multi-device gradients.
    init_checkpoint: a checkpoint for fine tuning.
    callbacks: a list of callbacks to call during training.
    n_weights_shards: shard weights into this many devices.
    additional_train_tasks: additional tasks which should be performed during
      training.
    additional_eval_tasks: additional tasks which should be performed during
      evaluation.
    additional_eval_streams: List[NamedStream], additional data streams that
      should be used during evaluation. Can be provided independently of
      additional_eval_tasks.

  Returns:
    trax.TrainerState or training.Loop if use_loop is True
  """
  base.N_WEIGHTS_SHARDS = n_weights_shards
  if (permanent_checkpoint_frequency is not None
      and permanent_checkpoints_at is not None):
    raise ValueError('Only one of ["permanent_checkpoint_frequency", '
                     '"permanent_checkpoints_at"] should be set.')
  if use_loop:
    n_devices = num_devices() or fastmath.local_device_count()

    # Prepare the training task.
    # Inputs is either an Inputs instance or a function that returns it.
    if callable(inputs):  # If we pass a function, e.g., through gin, call it.
      inputs = inputs()
    opt = optimizer if use_memory_efficient_trainer else optimizer()
    train_task = training.TrainTask(
        inputs.train_stream(n_devices),
        loss_layer=loss_fn,
        optimizer=opt,
        lr_schedule=lr_schedule_fn(),
        n_steps_per_checkpoint=eval_frequency,
        n_steps_per_permanent_checkpoint=permanent_checkpoint_frequency)

    if additional_train_tasks is None:
      additional_train_tasks = []

    # Prepare the evaluation.
    metrics_dict = metrics if metrics is not None else _DEFAULT_METRICS
    names, metrics = zip(*metrics_dict.items())
    eval_task = training.EvalTask(inputs.eval_stream(n_devices),
                                  metrics,
                                  metric_names=names,
                                  n_eval_batches=eval_steps)

    if additional_eval_tasks is None:
      additional_eval_tasks = []

    additional_eval_tasks_from_streams = []
    if additional_eval_streams is not None:
      for stream in additional_eval_streams:
        additional_eval_tasks_from_streams.append(
            training.EvalTask(stream.stream,
                              metrics,
                              metric_names=names,
                              n_eval_batches=eval_steps,
                              export_prefix=stream.name))

    # Prepare the training loop.
    checkpoint_at = None
    if checkpoints_at is not None:
      checkpoint_at = lambda step: step in checkpoints_at
    permanent_checkpoint_at = None
    if permanent_checkpoints_at is not None:
      permanent_checkpoint_at = (lambda step: step in permanent_checkpoints_at)

    # Setup the model.
    model_train = model(mode='train')
    model_predict_eval = model(mode='eval')
    if init_checkpoint:
      model_train.init_from_file(init_checkpoint, weights_only=True)
      model_predict_eval.init_from_file(init_checkpoint, weights_only=True)
    loop = training.Loop(
        model_train, [train_task] + additional_train_tasks,
        eval_model=model_predict_eval,
        eval_tasks=[eval_task] +
        additional_eval_tasks + additional_eval_tasks_from_streams,
        output_dir=output_dir,
        checkpoint_at=checkpoint_at,
        checkpoint_low_metric=checkpoint_lowest,
        checkpoint_high_metric=checkpoint_highest,
        permanent_checkpoint_at=permanent_checkpoint_at,
        n_devices=n_devices,
        loss_chunk_size=loss_chunk_size,
        use_memory_efficient_trainer=use_memory_efficient_trainer,
        adasum=adasum,
        random_seed=random_seed,
        callbacks=callbacks,
    )

    steps_to_go = steps - loop.step
    if steps_to_go <= 0:
      log('Stop training, already reached the total training steps %d' % steps)
      return loop

    # Train and return the loop.
    loop.run(steps_to_go)
    return loop

  n_devices = num_devices()
  trainer = trainer_class(model, loss_fn, optimizer, lr_schedule_fn(), inputs,
                          output_dir,
                          random_seed=random_seed,
                          n_devices=n_devices,
                          checkpoints_at=checkpoints_at,
                          metrics=metrics,
                          checkpoint_lowest=checkpoint_lowest,
                          checkpoint_highest=checkpoint_highest,
                          init_checkpoint=init_checkpoint)

  epoch_steps = [steps]  # Only training if eval_frequency is 0 or None
  if eval_frequency and eval_steps > 0:
    epoch_steps = itertools.chain([1,  # first epoch only 1 step
                                   eval_frequency - 1],
                                  itertools.repeat(eval_frequency))
  trainer.log_step('Starting training using %d devices' % trainer.n_devices)
  trainer.print_n_weights()

  try:
    for epoch_steps in epochs(steps, trainer.step, epoch_steps):
      trainer.train_epoch(epoch_steps, eval_steps)

      # Bookkeeping we do at the first step
      if trainer.step == 1:
        # Save computation graph (single-device only for now)
        if (save_graphs and fastmath.is_backend(fastmath.Backend.JAX)):
          trainer.save_computation_graphs()

        # Save Gin config
        trainer.save_gin()

    trainer.log_step('Training done')
  except Exception as e:
    raise e
  finally:
    trainer.close()
  return trainer.state


@gin.configurable
def num_devices(value=None):
  """Returns how many devices to use (if None, default, use all available)."""
  return value


@gin.configurable
def _jit_update_fn(predict_fn, loss_fn, optimizer, n_devices, jit=True):
  """Returns a (JIT-compiled) function that computes updates for one step."""
  model_and_loss = tl.Serial(predict_fn, loss_fn)
  # Gradients are always wrt. the first argument, so putting weights first.
  def model_and_loss_call(weights, batch, state, rng):
    res = model_and_loss(batch, weights=weights, state=state, rng=rng)
    return res, model_and_loss.state
  if n_devices == 1:  # TODO(lukaszkaiser): remove branch when not needed.
    def single_update(weights_and_slots, i, opt_params, batch, state, rng):
      weights, slots = weights_and_slots
      rng, subrng = jax_random.split(rng[0])
      grad_fn = fastmath.grad(model_and_loss_call, has_aux=True)
      grads, state = grad_fn(weights, batch, state, rng)
      new_weights, new_slots, stats = optimizer.tree_update(
          i, grads, weights, slots, opt_params)
      return (new_weights, new_slots), stats, state, [subrng]
    if jit:
      # TODO(lukaszkaiser): donate_argnums=(0,) when XLA supports it on GPU
      return fastmath.jit(single_update)
    else:
      return single_update

  # Else, for n_devices > 1:
  @functools.partial(fastmath.pmap, axis_name='batch')  # donate_argnums=(0,))
  def mapped_update(weights_and_slots, i, opt_params, batch, state, rng):
    """This is a multi-device version of the update function above."""
    # We assume all tensors have the first dimension = n_devices.
    weights, slots = weights_and_slots
    rng, subrng = jax_random.split(rng)
    grad_fn = fastmath.grad(model_and_loss_call, has_aux=True)
    grads, state = grad_fn(weights, batch, state, rng)
    # We do a psum(1.0) here instead of `n_devices` since `n_devices` is just
    # the number of devices on this host machine, however psum goes over all
    # devices of all hosts (ex: a TPU pod) and we need to be averaging over all
    # of them.
    #
    # Collect all gradients.
    grads = fastmath.psum(grads, 'batch')
    n_devices_total = fastmath.psum(np.array(1.0), 'batch')
    # Average across hosts.
    grads = jax.tree_util.tree_map(lambda g: g / n_devices_total, grads)

    new_weights, new_slots, stats = optimizer.tree_update(
        i, grads, weights, slots, opt_params)
    return (new_weights, new_slots), stats, state, subrng

  def update(weights_and_slots, i, opt_params, batch, state, rng):
    return mapped_update(weights_and_slots, np.repeat(i, n_devices),
                         opt_params, batch, state, rng)

  return update


@gin.configurable
def _jit_predict_fn(model_predict, metric_fn, n_devices, jit=True):
  """Returns a JIT-compiled predict function (unless jit=False)."""
  model = tl.Serial(model_predict, metric_fn)
  if not jit:
    return model.pure_fn

  return tl.jit_forward(model.pure_fn, n_devices)


@gin.configurable
def _jit_compute_loss_fn(predict_fn, loss_fn, n_devices, jit=True):
  """Returns a (JIT-compiled) function that computes the loss for one step."""
  if n_devices == 1:  # TODO(lukaszkaiser): remove branch when not needed.
    def single_compute_loss(opt_state, batch, state, rng):
      rng, subrng = jax_random.split(rng[0])
      loss_val, state = loss_fn(opt_state[0], batch, predict_fn, state, rng)
      return loss_val, state, [subrng]
    return fastmath.jit(single_compute_loss) if jit else single_compute_loss

  # Else, for n_devices > 1:
  @functools.partial(fastmath.pmap, axis_name='batch')
  def mapped_compute_loss(opt_state, batch, state, rng):
    """This is a multi-device version of the update function above."""
    # We assume all tensors have the first dimension = n_devices.
    rng, subrng = jax_random.split(rng)
    loss_val, state = loss_fn(opt_state[0], batch, predict_fn, state, rng)
    return loss_val, state, subrng

  def compute_loss(opt_state, batch, state, rng):
    return mapped_compute_loss(
        opt_state, _reshape_by_device(batch, n_devices), state, rng)

  return compute_loss


def log(s, stdout=True):
  logging.info(s)
  if stdout:
    print(s)
    sys.stdout.flush()


def epochs(total_steps, steps_to_skip, epoch_steps):
  """Generates the number of steps in each epoch before reaching total_steps.

  Args:
    total_steps: int, total number of steps.
    steps_to_skip: int, number of steps to skip because of a restart.
    epoch_steps: iterable of int, numbers of steps in each epoch.

  Yields:
    epoch_steps: int, number of steps in this epoch
  """
  steps_to_go = total_steps - steps_to_skip
  epoch_steps = iter(epoch_steps)

  # Remove the desired number of steps from the stream.
  for steps_this_epoch in epoch_steps:
    if steps_this_epoch > steps_to_skip:
      # Put back the number of steps left in the unfinished epoch.
      epoch_steps = itertools.chain(
          [steps_this_epoch - steps_to_skip], epoch_steps)
    if steps_this_epoch >= steps_to_skip:
      break
    steps_to_skip -= steps_this_epoch

  # Yield the remaining steps per epoch up to total_steps.
  for steps_this_epoch in epoch_steps:
    steps_this_epoch = min(steps_this_epoch, steps_to_go)
    yield steps_this_epoch
    steps_to_go -= steps_this_epoch
    if steps_to_go == 0:
      break


def make_trainer_state_dict(step,
                            opt_state,
                            history,
                            model_state,
                            input_signature):
  """Creates a trainer state dictionary to save to disk.

  Args:
    step: int, a step number
    opt_state: OptState namedtuple
    history: `trax.history.History`, the history object.
    model_state: A nested structure of the model state.
    input_signature: signature of model inputs.

  Returns:
    A dictionary with the fields of TrainerState and OptState flattened.
  """
  flat_weights, flat_state = tl.flatten_weights_and_state(
      opt_state.weights, model_state)
  return {
      'step': step,
      'flat_weights': flat_weights,
      'slots': opt_state.slots,
      'opt_params': opt_state.opt_params,
      'history': history,
      'flat_state': flat_state,
      'input_signature': input_signature,
      'version_timestamp': 'Jun-18-2020'  # To update in the future if needed.
  }


def trainer_state_from_dict(trainer_state_dict, model):
  """Given the trainer state dictionary, returns `TrainerState`."""
  # TODO(afrozm): This becomes simpler if OptState is flattened into
  # TrainerState.
  step = trainer_state_dict['step']
  history = trainer_state_dict['history']
  input_signature = trainer_state_dict['input_signature']
  weights_and_state_sig = model.weights_and_state_signature(input_signature)
  weights, model_state = tl.unflatten_weights_and_state(
      trainer_state_dict['flat_weights'], trainer_state_dict['flat_state'],
      weights_and_state_sig)
  opt_state = OptState(
      weights=weights,
      slots=trainer_state_dict['slots'],
      opt_params=trainer_state_dict['opt_params'])
  return TrainerState(step=step, opt_state=OptState(*opt_state),
                      history=history, model_state=model_state)


def load_trainer_state(output_dir, model, weights_file=None):
  """Returns a TrainerState instance loaded from the given `output_dir`."""
  if weights_file is None:
    weights_file = os.path.join(output_dir, 'model.pkl.gz')
    if not tf.io.gfile.exists(weights_file):
      return TrainerState(step=None, opt_state=None,
                          history=trax_history.History(), model_state=None)
  elif not tf.io.gfile.exists(weights_file):
    raise ValueError('File not found: %s' % weights_file)

  trainer_state_dict = training.unpickle_from_file(weights_file, gzip=True)
  trainer_state = trainer_state_from_dict(trainer_state_dict, model)
  log('Model loaded from %s at step %d' % (weights_file, trainer_state.step))
  logging.debug('From loaded model : history = %s', trainer_state.history)
  return trainer_state


def _reshape_by_device(x, n_devices):
  """Reshapes possibly nested x into a shape (n_devices, ...)."""
  return tl.reshape_by_device(x, n_devices)  # pylint: disable=protected-access


def _nested_reduce(f, x):
  """Fold the function f to the nested structure x (dicts, tuples, lists)."""
  if isinstance(x, list):
    return f([_nested_reduce(f, y) for y in x])
  if isinstance(x, tuple):
    return f([_nested_reduce(f, y) for y in x])
  if isinstance(x, dict):
    return f([_nested_reduce(f, v) for (_, v) in x.items()])
  return x


def _sizes(x):
  """Get a structure of sizes for a structure of nested arrays."""
  def size(x):
    try:
      return x.size
    except Exception:  # pylint: disable=broad-except
      return 0
  return fastmath.nested_map(size, x)


def _repeat_stream(stream, n_devices):
  """Repeat a stream indefinitely."""
  while True:
    for example in stream(n_devices):
      yield example



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Core class and functions for handling data abstractly as shapes/dtypes."""

import numpy as np
import tensorflow as tf


class ShapeDtype:
  """A NumPy ndarray-like object abstracted as shape and dtype.

  Main use is for representing input and output signatures.
  """
  __slots__ = ['shape', 'dtype']

  def __init__(self, shape, dtype=np.float32):
    """Creates a `ShapeDtype` instance, with canonicalized `shape` and `dtype`.

    Args:
      shape: A tuple or list, each element of which is an int or, less often,
          `None`.
      dtype: A `dtype` object, either from NumPy or TensorFlow.

    Returns:
      A `ShapeDtype` instance whose `shape` is a tuple and `dtype` is a NumPy
      `dtype` object.
    """
    # Canonicalize shape and dtype.
    if isinstance(shape, tf.TensorShape):
      shape = shape.as_list()
    if isinstance(shape, list):
      shape = tuple(shape)
    if not isinstance(shape, tuple):
      raise TypeError('shape must be tuple or list; got: {}'.format(shape))
    if isinstance(dtype, tf.DType):
      dtype = dtype.as_numpy_dtype

    self.shape = shape
    self.dtype = dtype

  def __eq__(self, other):
    return (isinstance(other, self.__class__)
            and self.shape == other.shape
            and self.dtype == other.dtype)

  def __ne__(self, other):
    return not self == other

  def __repr__(self):
    return 'ShapeDtype{{shape:{}, dtype:{}}}'.format(self.shape, self.dtype)

  def __len__(self):
    """Returns length of 1; relevant to input and output signatures."""
    return 1

  def as_tuple(self):
    return self.shape, self.dtype

  def replace(self, **kwargs):
    """Creates a copy of the object with some parameters replaced."""
    return type(self)(
        shape=kwargs.pop('shape', self.shape),
        dtype=kwargs.pop('dtype', self.dtype),
    )


def signature(obj):
  """Returns a `ShapeDtype` signature for the given `obj`.

  A signature is either a `ShapeDtype` instance or a tuple of `ShapeDtype`
  instances. Note that this function is permissive with respect to its inputs
  (accepts lists or tuples or dicts, and underlying objects can be any type
  as long as they have shape and dtype attributes) and returns the corresponding
  nested structure of `ShapeDtype`.

  Args:
    obj: An object that has `shape` and `dtype` attributes, or a list/tuple/dict
        of such objects.

  Returns:
    A corresponding nested structure of `ShapeDtype` instances.
  """
  if isinstance(obj, (list, tuple)):
    output = tuple(signature(x) for x in obj)
    return output if isinstance(obj, tuple) else list(output)
  elif isinstance(obj, dict):
    return {k: signature(v) for (k, v) in obj.items()}
  else:
    return ShapeDtype(obj.shape, obj.dtype)


def splice_signatures(*sigs):
  """Creates a new signature by splicing together any number of signatures.

  The splicing effectively flattens the top level input signatures. For
  instance, it would perform the following mapping:

    - `*sigs: sd1, (sd2, sd3, sd4), (), sd5`
    - return: `(sd1, sd2, sd3, sd4, sd5)`

  Args:
    *sigs: Any number of signatures. A signature is either a `ShapeDtype`
        instance or a tuple of `ShapeDtype` instances.

  Returns:
    A single `ShapeDtype` instance if the spliced signature has one element,
    else a tuple of `ShapeDtype` instances.
  """
  result_sigs = []
  for sig in sigs:
    if isinstance(sig, (list, tuple)):
      result_sigs.extend(sig)
    else:
      result_sigs.append(sig)
  return result_sigs[0] if len(result_sigs) == 1 else tuple(result_sigs)


def assert_shape_equals(array, shape):
  """Asserts that an array has the given shape."""
  assert array.shape == shape, (
      'Invalid shape {}; expected {}.'.format(array.shape, shape)
  )


def assert_same_shape(array1, array2):
  """Asserts that two arrays have the same shapes."""
  assert_shape_equals(array1, array2.shape)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Rotary positional embeddings.

  Rotary positional embedding implementation, as described in:
  https://arxiv.org/pdf/2104.09864.pdf
"""

# from trax import layers as tl
from trax.fastmath import numpy as jnp
from trax.layers import core


def rotate(x):
  """Rotate function."""
  _, l, d = x.shape
  inv_freq = jnp.exp(jnp.arange(0, d, 2) * -(jnp.log(10000.0) / d))
  positions = jnp.arange(l)
  freqs = jnp.einsum('i,j->ij', positions, inv_freq)
  emb = jnp.concatenate((freqs, freqs), axis=-1)
  cos = jnp.cos(emb)
  sin = jnp.sin(emb)

  def mul(vecs, pos_emb):
    return jnp.einsum('bld,ld->bld', vecs, pos_emb)

  def rotate_half(x):
    x1, x2 = x[..., :x.shape[-1] // 2], x[..., x.shape[-1] // 2:]
    return jnp.concatenate((-x2, x1), axis=x1.ndim - 1)

  return mul(x, cos) + mul(rotate_half(x), sin)


def Rotate():  # pylint: disable=invalid-name
  return core.Fn('Rotate', rotate)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for trax.layers.research.efficient_attention."""

from absl.testing import parameterized
import jax
import numpy as np
from tensorflow import test

from trax import fastmath
from trax import shapes
from trax.fastmath import numpy as jnp
from trax.layers.research import efficient_attention


class EfficientAttentionTest(test.TestCase, parameterized.TestCase):

  def test_self_attention(self):
    with fastmath.use_backend(fastmath.Backend.JAX):
      layer = efficient_attention.SelfAttention(
          n_heads=5, d_qk=7, d_v=17, share_qk=False, causal=True,
          chunk_len=8, n_chunks_before=1, n_chunks_after=0,
          use_reference_code=True, attention_dropout=0.0, mode='train')
      x = np.ones((3, 32, 8)).astype(np.float32)
      _, _ = layer.init(shapes.signature(x))
      y = layer(x)
      self.assertEqual(y.shape, x.shape)

  def test_lsh_ff(self):
    with fastmath.use_backend(fastmath.Backend.JAX):
      layer = efficient_attention.LSHFF(d_ff=1024*8, n_buckets=[16, 8])
      x = np.ones((3, 7, 1024)).astype(np.float32)
      _, _ = layer.init(shapes.signature(x))
      y = layer(x)
      self.assertEqual(y.shape, x.shape)

  def test_self_attention_tf(self):
    with fastmath.use_backend(fastmath.Backend.TFNP):
      layer = efficient_attention.SelfAttention(
          n_heads=5, d_qk=7, d_v=17, share_qk=False, causal=True,
          chunk_len=8, n_chunks_before=1, n_chunks_after=0,
          use_reference_code=True, attention_dropout=0.0, mode='train')
      x = np.ones((3, 32, 8)).astype(np.float32)
      _, _ = layer.init(shapes.signature(x))
      y = layer(x)
      self.assertEqual(y.shape, x.shape)

  def test_lsh_self_attention(self):
    with fastmath.use_backend(fastmath.Backend.JAX):
      layer = efficient_attention.LSHSelfAttention(
          n_heads=5, d_qk=7, d_v=17, causal=True,
          chunk_len=8, n_chunks_before=1, n_chunks_after=0,
          n_hashes=2, n_buckets=4,
          use_reference_code=True, attention_dropout=0.0, mode='train')
      x = np.ones((3, 32, 8)).astype(np.float32)
      _, _ = layer.init(shapes.signature(x))
      y = layer(x)
      self.assertEqual(y.shape, x.shape)


  def _run_forward_and_backward(self, model, inp, weights, state):
    def forward(inp, weights):
      return model.pure_fn(
          inp, weights, state, rng=jax.random.PRNGKey(0))
    out, vjpfun, new_state = jax.vjp(forward, inp, weights, has_aux=True)
    inp_grad, weights_grad = vjpfun(fastmath.numpy.ones_like(inp))
    return out, new_state, inp_grad, weights_grad

  def _test_equivalence_to_reference_code(
      self, model_cls, inp, input_signature, common_kwargs, *test_kwargs):
    ref_model = model_cls(use_reference_code=True, **common_kwargs)
    rng = fastmath.random.get_prng(123)
    weights, state = ref_model.init(input_signature, rng)

    ref_all = self._run_forward_and_backward(ref_model, inp, weights, state)
    ref_out, ref_state, ref_inp_grad, ref_weights_grad = ref_all

    for kwargs in test_kwargs:
      test_model = model_cls(**common_kwargs, **kwargs)
      state = test_model.init(input_signature, rng)[1]
      test_all = self._run_forward_and_backward(test_model, inp, weights, state)
      test_out, test_state, test_inp_grad, test_weights_grad = test_all

      self.assertEqual(jax.tree_structure(ref_out),
                       jax.tree_structure(test_out))
      self.assertEqual(jax.tree_structure(ref_state),
                       jax.tree_structure(test_state))
      self.assertEqual(jax.tree_structure(ref_inp_grad),
                       jax.tree_structure(test_inp_grad))
      self.assertEqual(jax.tree_structure(ref_weights_grad),
                       jax.tree_structure(test_weights_grad))

      check_close = lambda x, y: self.assertAllClose(x, y, rtol=2e-3, atol=2e-3)
      fastmath.nested_map_multiarg(check_close, ref_out, test_out)
      fastmath.nested_map_multiarg(check_close, ref_state, test_state)
      fastmath.nested_map_multiarg(check_close, ref_inp_grad, test_inp_grad)
      fastmath.nested_map_multiarg(check_close, ref_weights_grad,
                                   test_weights_grad)

  def test_batching_self_attention(self):
    with fastmath.use_backend(fastmath.Backend.JAX):
      common_kwargs = dict(
          n_heads=6, d_qk=7, d_v=17, share_qk=False, causal=True,
          chunk_len=5, n_chunks_before=1, n_chunks_after=0,
          attention_dropout=0.2, output_dropout=0.1, mode='train',
      )
      test_kwargs = []
      for n_parallel_heads in [1, 3, 6, 12]:
        for use_python_loop in [True, False]:
          test_kwargs.append(dict(n_parallel_heads=n_parallel_heads,
                                  use_python_loop=use_python_loop))

      x = jax.random.uniform(
          jax.random.PRNGKey(0), (2, 10, 13), dtype=jnp.float32)
      input_signature = shapes.signature(x)
      self._test_equivalence_to_reference_code(
          efficient_attention.SelfAttention,
          x, input_signature,
          common_kwargs, *test_kwargs)

  def test_batching_lsh_self_attention(self):
    with fastmath.use_backend(fastmath.Backend.JAX):
      common_kwargs = dict(
          n_heads=6, d_qk=7, d_v=17, causal=True,
          chunk_len=5, n_chunks_before=1, n_chunks_after=0,
          n_hashes=2, n_buckets=4,
          attention_dropout=0.2, output_dropout=0.1, mode='train',
      )
      test_kwargs = []
      for n_parallel_heads in [1, 3, 6, 12]:
        for use_python_loop in [True, False]:
          test_kwargs.append(dict(n_parallel_heads=n_parallel_heads,
                                  use_python_loop=use_python_loop))

      x = jax.random.uniform(
          jax.random.PRNGKey(0), (2, 10, 13), dtype=jnp.float32)
      input_signature = shapes.signature(x)
      self._test_equivalence_to_reference_code(
          efficient_attention.LSHSelfAttention,
          x, input_signature,
          common_kwargs, *test_kwargs)

  def _test_fast_inference(
      self, model_cls, x, input_signature, common_kwargs, *test_kwargs):
    ref_model = model_cls(use_reference_code=True, mode='eval', **common_kwargs)
    weights, state = ref_model.init(input_signature)

    ref_out, _ = ref_model.pure_fn(
        x, weights, state, rng=jax.random.PRNGKey(0))

    def get_slice(pytree, i):
      def get_slice_for_val(x):
        if isinstance(x, shapes.ShapeDtype):
          return shapes.ShapeDtype(shape=x.shape[:1] + (1,) + x.shape[2:],
                                   dtype=x.dtype)
        else:
          return x[:, i:i+1]
      return jax.tree_map(get_slice_for_val, pytree)

    seqlen = x[0].shape[1] if isinstance(x, (tuple, list)) else x.shape[1]

    for kwargs in test_kwargs:
      test_model = model_cls(mode='predict', **common_kwargs, **kwargs)
      cur_state = test_model.init(get_slice(input_signature, 0))[1]
      out = []
      for i in range(seqlen):
        cur_out, cur_state = test_model.pure_fn(
            get_slice(x, i), weights, cur_state, jax.random.PRNGKey(0))
        out.append(cur_out)
      out = jnp.concatenate(out, axis=1)

      self.assertAllClose(out, ref_out, rtol=1e-3, atol=1e-3)

  def test_fast_inference_self_attention(self):
    with fastmath.use_backend(fastmath.Backend.JAX):
      common_kwargs = dict(
          n_heads=6, d_qk=7, d_v=17, share_qk=False, causal=True,
          chunk_len=5, n_chunks_before=1, n_chunks_after=0,
          attention_dropout=0.0, output_dropout=0.0,
      )
      test_kwargs = []
      for n_parallel_heads in [1, 3, 6, 12]:
        for use_python_loop in [True, False]:
          test_kwargs.append(dict(n_parallel_heads=n_parallel_heads,
                                  use_python_loop=use_python_loop))

      x = jax.random.uniform(
          jax.random.PRNGKey(0), (2, 10, 13), dtype=jnp.float32)
      input_signature = shapes.signature(x)
      self._test_fast_inference(
          efficient_attention.SelfAttention,
          x, input_signature,
          common_kwargs, *test_kwargs)

  def _test_lsh_self_attention_deterministic_given_seed(self, causal=False):
    # Once the initialization and the call seeds are pinned down we have
    # deterministic output.
    with fastmath.use_backend(fastmath.Backend.JAX):
      layer = efficient_attention.LSHSelfAttention(
          n_heads=5, d_qk=7, d_v=17, causal=causal,
          chunk_len=8, n_chunks_before=1, n_chunks_after=0,
          n_hashes=2, n_buckets=4,
          use_reference_code=True, attention_dropout=0.0, mode='train')
      x = np.ones((3, 32, 8)).astype(np.float32)

      def get_output():
        _, _ = layer.init(shapes.signature(x), jax.random.PRNGKey(0))
        return layer(x, rng=jax.random.PRNGKey(1))

      ys = [get_output() for _ in range(10)]

      self.assertEqual(ys[0].shape, x.shape)

      for y in ys[1:]:
        np.testing.assert_array_almost_equal(ys[0], y, decimal=6)

  def test_lsh_determinism_causal(self):
    self._test_lsh_self_attention_deterministic_given_seed(causal=True)

  def test_lsh_determinism_non_causal(self):
    self._test_lsh_self_attention_deterministic_given_seed(causal=False)

  def test_lsh_self_attention_masked_non_causal(self):
    # Test that when the input that is in the masked area changes the attention
    # for the un-masked outputs doesn't change, but the masked region does
    # change.
    with fastmath.use_backend(fastmath.Backend.JAX):
      layer = efficient_attention.LSHSelfAttention(
          n_heads=5, d_qk=7, d_v=17, causal=False, masked=True,
          chunk_len=8, n_chunks_before=1, n_chunks_after=0,
          n_hashes=2, n_buckets=4,
          use_reference_code=True, attention_dropout=0.0, mode='train')

      batch = 5
      max_len = 32
      hidden = 8

      x = np.random.uniform(size=(batch, max_len, hidden))
      mask = np.ones((batch, max_len)).astype(np.bool)
      rngs = jax.random.randint(
          jax.random.PRNGKey(0), (batch,), minval=1, maxval=max_len - 1)

      # Set some suffix of each mask[b] to 0.
      for i in range(batch):
        mask[i, rngs[i]:] = 0

      # Fix rngs and get the output for the LSH layer.
      def get_output(x, mask):
        xs = [x, mask]
        _, _ = layer.init(shapes.signature(xs), jax.random.PRNGKey(0))
        return layer(xs, rng=jax.random.PRNGKey(1))

      # Get the attention output for masked x.
      y = get_output(x, mask)

      # Change x, but only in the masked regions.
      for i in range(batch):
        x[i, rngs[i]:] = np.random.uniform(size=(max_len - rngs[i], hidden))

      y2 = get_output(x, mask)

      for i in range(batch):
        # y and y2 should be identical in the non-masked part.
        np.testing.assert_array_almost_equal(y[i, :rngs[i]], y2[i, :rngs[i]],
                                             decimal=6)

        # In the masked out part, they should be different.
        self.assertGreater(
            np.mean(np.abs(y[i, rngs[i]:] - y2[i, rngs[i]:])), 1e-5)

  @parameterized.named_parameters(('_weights_2', 2), ('_weights_3', 3))
  def test_pure_lsh_wrapper_causal_non_masked(self, num_weights):
    with fastmath.use_backend(fastmath.Backend.JAX):
      n_heads = 5
      batch, seqlen, d_head = 3, 32, 8
      n_hashes = 2
      d_model = n_heads * d_head
      layer = efficient_attention.PureLSHSelfAttentionWrapper(
          n_heads=n_heads, d_qk=d_head, d_v=d_head, causal=True, masked=False,
          chunk_len=8, n_chunks_before=1, n_chunks_after=0,
          n_hashes=n_hashes, n_buckets=4, bias=False,
          pure_lsh_implementation=efficient_attention.PureLSHSelfAttention,
          mode='train', num_weights=num_weights)

      rng = jax.random.PRNGKey(0)
      rng, x_rng = jax.random.split(rng)

      input_shape = (batch, seqlen, d_model)
      x = jax.random.uniform(x_rng, input_shape, dtype=jnp.float32)

      inp = x
      w, s = layer.init(shapes.signature(inp))
      o = layer(inp)

      # Get the actual weights.
      weights = fastmath.tree_leaves(w)
      # Assert number of weights is as expected, the extra 1 is for output.
      self.assertLen(weights, num_weights + 1)

      # Assert each weight is of the expected shape.
      for i in range(num_weights + 1):
        self.assertEqual(weights[i].shape, (d_model, d_model))

      # Test that the output and the input shape match.
      self.assertEqual(inp.shape, o.shape)

      # Assert state is the shape expected.
      state = fastmath.tree_leaves(s)
      self.assertLen(state, 2)
      # buckets
      self.assertEqual(state[0].shape, (batch * n_heads, n_hashes * seqlen))
      # rngs
      self.assertEqual(state[1].shape, (batch * n_heads, 2))

  @parameterized.named_parameters(('_weights_2', 2), ('_weights_3', 3))
  def test_pure_lsh_wrapper_non_causal_masked(self, num_weights):
    with fastmath.use_backend(fastmath.Backend.JAX):
      n_heads = 5
      batch, seqlen, d_head = 3, 32, 8
      num_weights = 2
      n_hashes = 2
      d_model = n_heads * d_head
      layer = efficient_attention.PureLSHSelfAttentionWrapper(
          n_heads=n_heads, d_qk=d_head, d_v=d_head, causal=False, masked=True,
          chunk_len=8, n_chunks_before=1, n_chunks_after=0,
          n_hashes=n_hashes, n_buckets=4, bias=False,
          pure_lsh_implementation=efficient_attention.PureLSHSelfAttention,
          mode='train', num_weights=num_weights)

      rng = jax.random.PRNGKey(0)
      rng, x_rng = jax.random.split(rng)

      input_shape = (batch, seqlen, d_model)
      x = jax.random.uniform(x_rng, input_shape, dtype=jnp.float32)
      mask = jnp.ones((batch, seqlen), dtype=jnp.int32)

      inp = (x, mask)
      w, s = layer.init(shapes.signature(inp))
      o = layer(inp)

      # Get the actual weights.
      weights = fastmath.tree_leaves(w)
      # Assert number of weights is as expected, the extra 1 is for output.
      self.assertLen(weights, num_weights + 1)

      # Assert each weight is of the expected shape.
      for i in range(num_weights + 1):
        self.assertEqual(weights[i].shape, (d_model, d_model))

      # Test that the output and the x's shape match.
      self.assertEqual(x.shape, o.shape)

      # Assert state is the shape expected.
      state = fastmath.tree_leaves(s)
      self.assertLen(state, 2)
      # buckets
      self.assertEqual(state[0].shape, (batch * n_heads, n_hashes * seqlen))
      # rngs
      self.assertEqual(state[1].shape, (batch * n_heads, 2))

  def test_lsh_and_pure_lsh_self_attention_equivalence(self):
    # Given the same weight matrices and random numbers, do these produce the
    # same output.
    with fastmath.use_backend(fastmath.Backend.JAX):
      n_heads = 4
      d_head = 4
      d_model = n_heads * d_head
      pure_lsh_layer = efficient_attention.PureLSHSelfAttention(
          n_heads=n_heads, d_qk=d_head, d_v=d_head, causal=True, masked=False,
          chunk_len=8, n_chunks_before=1, n_chunks_after=0,
          n_hashes=4, n_buckets=8,
          use_reference_code=False,
          attention_dropout=0.0,
          use_python_loop=True,
          bias=False, mode='train')
      lsh_layer = efficient_attention.LSHSelfAttention(
          n_heads=n_heads, d_qk=d_head, d_v=d_head, causal=True, masked=False,
          chunk_len=8, n_chunks_before=1, n_chunks_after=0,
          n_hashes=4, n_buckets=8,
          use_reference_code=False,
          attention_dropout=0.0,
          use_python_loop=True,
          mode='train')

      batch, seqlen = 3, 32
      input_shape = (batch, seqlen, d_model)

      x = jax.random.uniform(jax.random.PRNGKey(0), input_shape,
                             dtype=jnp.float32)
      lsh_layer_input = x

      call_rng = jax.random.PRNGKey(42)

      lsh_layer_weights, lsh_layer_state = lsh_layer.init(
          shapes.signature(lsh_layer_input))
      lsh_layer.rng = call_rng
      lsh_layer_output = lsh_layer(lsh_layer_input)

      # Shapes are: (n_heads, d_model, d_head), (n_heads, d_model, d_head),
      # (n_heads, d_head, d_model)
      # Abbreviated as - hmn, hmn, hnm
      w_qk, w_v, w_o = lsh_layer_weights

      qk = jnp.einsum('blm,hmn->bhln', x, w_qk)
      qk = qk.reshape((-1, qk.shape[2], qk.shape[3]))

      v = jnp.einsum('blm,hmn->bhln', x, w_v)
      v = v.reshape((-1, v.shape[2], v.shape[3]))

      pure_lsh_layer_input = (qk, v)
      _, _ = pure_lsh_layer.init(shapes.signature(pure_lsh_layer_input))
      pure_lsh_layer.rng = call_rng
      pure_lsh_layer.state = lsh_layer_state
      pure_lsh_layer_output = pure_lsh_layer(pure_lsh_layer_input)

      # b*h,l,n
      pure_lsh_layer_output = pure_lsh_layer_output.reshape(
          (batch, -1) + pure_lsh_layer_output.shape[1:])
      pure_lsh_layer_output_projected = (
          jnp.einsum('bhld,hdm->blm', pure_lsh_layer_output, w_o))

      diff = pure_lsh_layer_output_projected - lsh_layer_output
      avg_diff = jnp.sum(jnp.abs(diff)) / jnp.sum(jnp.ones_like(diff))

      self.assertLess(avg_diff, 1e-5)

if __name__ == '__main__':
  test.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Test training an MNIST model 100 steps (saves time vs. 2000 steps)."""

import io
import itertools
from unittest import mock

from absl.testing import absltest

from trax import layers as tl
from trax.data import inputs
from trax.data import tf_inputs
from trax.optimizers import adam
from trax.supervised import training


class MnistTest(absltest.TestCase):

  @mock.patch('sys.stdout', new_callable=io.StringIO)
  def test_train_mnist_single_task(self, mock_stdout):
    """Train MNIST model a bit, to compare to other implementations."""
    mnist_model = _build_model(two_heads=False)
    (task, eval_task) = _mnist_tasks()
    training_session = training.Loop(
        mnist_model,
        tasks=[task],
        eval_tasks=[eval_task],
        eval_at=lambda step_n: step_n % 20 == 0,
    )

    training_session.run(n_steps=100)
    self.assertEqual(training_session.step, 100)

    # Assert that we reach at least 80% eval accuracy.
    self.assertGreater(_read_metric('WeightedCategoryAccuracy', mock_stdout),
                       0.8)

  @mock.patch('sys.stdout', new_callable=io.StringIO)
  def test_train_mnist_multitask(self, mock_stdout):
    """Train two-head MNIST model a bit, to compare to other implementations."""
    mnist_model = _build_model(two_heads=True)
    # MNIST classification task.
    (cls_task, cls_eval_task) = _mnist_tasks(head=tl.Select([0], n_in=2))
    # Auxiliary brightness prediction task.
    reg_task = training.TrainTask(
        itertools.cycle(_mnist_brightness_dataset().train_stream(1)),
        tl.Serial(tl.Select([1]), tl.L2Loss()),
        adam.Adam(0.001),
    )
    reg_eval_task = training.EvalTask(
        itertools.cycle(_mnist_brightness_dataset().eval_stream(1)),
        [tl.Serial(tl.Select([1]), tl.L2Loss())],
        n_eval_batches=1,
        metric_names=['L2'],
    )
    training_session = training.Loop(
        mnist_model,
        tasks=[cls_task, reg_task],
        eval_tasks=[cls_eval_task, reg_eval_task],
        eval_at=lambda step_n: step_n % 20 == 0,
        which_task=lambda step_n: step_n % 2,
    )

    training_session.run(n_steps=100)
    self.assertEqual(training_session.step, 100)

    # Assert that we reach at least 80% eval accuracy on MNIST.
    self.assertGreater(_read_metric('WeightedCategoryAccuracy', mock_stdout),
                       0.8)
    # Assert that we get below 0.03 brightness prediction error.
    self.assertLess(_read_metric('L2', mock_stdout), 0.03)


def _build_model(two_heads):
  cls_head = tl.Dense(10)
  if two_heads:
    reg_head = tl.Dense(1)
    heads = tl.Branch(cls_head, reg_head)
  else:
    heads = cls_head
  return tl.Serial(
      tl.Fn('ScaleInput', lambda x: x / 255),
      tl.Flatten(),
      tl.Dense(512),
      tl.Relu(),
      tl.Dense(512),
      tl.Relu(),
      heads,
  )


def _mnist_dataset():
  """Loads (and caches) the standard MNIST data set."""
  streams = tf_inputs.data_streams('mnist')
  return inputs.batcher(streams, variable_shapes=False,
                        batch_size_per_device=256,
                        eval_batch_size=256)


def _mnist_brightness_dataset():
  """Loads (and caches) a MNIST mean brightness data set."""
  def preprocess_stream(stream):
    def new_stream():
      for (image, _) in stream():
        yield (image, (image / 255).mean()[None])
    return new_stream

  streams = tuple(map(preprocess_stream, tf_inputs.data_streams('mnist')))
  return inputs.batcher(streams, variable_shapes=False,
                        batch_size_per_device=256,
                        eval_batch_size=256)


def _mnist_tasks(head=None):
  """Creates MNIST training and evaluation tasks.

  Args:
    head: Adaptor layer to put before loss and accuracy layers in the tasks.

  Returns:
    A pair (train_task, eval_task) consisting of the MNIST training task and the
    MNIST evaluation task using cross-entropy as loss and accuracy as metric.
  """
  loss = tl.WeightedCategoryCrossEntropy()
  accuracy = tl.WeightedCategoryAccuracy()
  if head is not None:
    loss = tl.Serial(head, loss)
    accuracy = tl.Serial(head, accuracy)
  task = training.TrainTask(
      itertools.cycle(_mnist_dataset().train_stream(1)),
      loss,
      adam.Adam(0.001),
  )
  eval_task = training.EvalTask(
      itertools.cycle(_mnist_dataset().eval_stream(1)),
      [loss, accuracy],
      n_eval_batches=10,
      metric_names=['CrossEntropy', 'WeightedCategoryAccuracy'],
  )
  return (task, eval_task)


def _read_metric(metric_name, stdout):
  log = stdout.getvalue()
  metric_log = [line for line in log.split('\n') if metric_name in line][-1]
  return float(metric_log.strip().split(' ')[-1])


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A few utilities for tests."""

import sys

from absl import flags

FLAGS = flags.FLAGS


# pytest doesn't run the test as a main, so it doesn't parse the flags
# so if flags are required in tests, this will ensure that flags are manually
# parsed and the desired flag exists.
def ensure_flag(flag_str):
  try:
    getattr(FLAGS, flag_str)
  except flags.UnparsedFlagAccessError:
    # Manually parse flags.
    FLAGS(sys.argv)
  finally:
    assert getattr(FLAGS, flag_str)
